# Rotinas de testes associadas ao arquivo MEDIAS-USIE.CSV do NWLISTOP
from inewave.nwlistop.mediasusie import Mediasusie


ARQ_TESTE = "tests/_arquivos/MEDIAS-USIE.CSV"


def test_eq_mediasusie():
    leitor = Mediasusie.read(ARQ_TESTE)
    leitor2 = Mediasusie.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (0, 2)



from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.datetimefield import DatetimeField
from datetime import datetime

from typing import Optional


class HidreletricaCurvaJusante(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-CURVAJUSANTE"
    IDENTIFIER_DIGITS = 26
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            FloatField(decimal_digits=4),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina hidrelétrica relacionada ao polinômio.

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def indice_familia(self) -> Optional[int]:
        """
        O índice da família de polinômios.

        :return: O índice (sequencial).
        :rtype: int | None
        """
        return self.data[1]

    @indice_familia.setter
    def indice_familia(self, c: int):
        self.data[1] = c

    @property
    def nivel_montante_referencia(self) -> Optional[float]:
        """
        O nível de montante da usina de jusante
        de referência.

        :return: O nível em metros
        :rtype: float | None
        """
        return self.data[2]

    @nivel_montante_referencia.setter
    def nivel_montante_referencia(self, c: float):
        self.data[2] = c


class HidreletricaCurvaJusantePolinomioPorPartes(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-CURVAJUSANTE-POLINOMIOPORPARTES"
    IDENTIFIER_DIGITS = 45
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina hidrelétrica relacionada ao polinômio.

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def indice_familia(self) -> Optional[int]:
        """
        O índice da família de polinômios.

        :return: O índice (sequencial).
        :rtype: int | None
        """
        return self.data[1]

    @indice_familia.setter
    def indice_familia(self, c: int):
        self.data[1] = c

    @property
    def numero_polinomios(self) -> Optional[int]:
        """
        O número de polinômios existentes na família.

        :return: O número de polinômios
        :rtype: int | None
        """
        return self.data[2]

    @numero_polinomios.setter
    def numero_polinomios(self, c: int):
        self.data[2] = c


class HidreletricaCurvaJusantePolinomioPorPartesSegmento(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-CURVAJUSANTE-POLINOMIOPORPARTES-SEGMENTO"
    IDENTIFIER_DIGITS = 54
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
            IntegerField(),
            FloatField(size=24, decimal_digits=3),
            FloatField(size=24, decimal_digits=3),
            FloatField(size=24, decimal_digits=14, format="E"),
            FloatField(size=24, decimal_digits=14, format="E"),
            FloatField(size=24, decimal_digits=14, format="E"),
            FloatField(size=24, decimal_digits=14, format="E"),
            FloatField(size=24, decimal_digits=14, format="E"),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina hidrelétrica relacionada ao polinômio.

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def indice_familia(self) -> Optional[int]:
        """
        O índice da família de polinômios.

        :return: O índice (sequencial).
        :rtype: int | None
        """
        return self.data[1]

    @indice_familia.setter
    def indice_familia(self, c: int):
        self.data[1] = c

    @property
    def indice_polinomio(self) -> Optional[int]:
        """
        O índice do polinômio da respectiva família.

        :return: O índice do polinômio
        :rtype: int | None
        """
        return self.data[2]

    @indice_polinomio.setter
    def indice_polinomio(self, c: int):
        self.data[2] = c

    @property
    def limite_inferior_vazao_jusante(self) -> Optional[float]:
        """
        O limite inferior de vazão de jusante (defluência mais lateral)
        para janela de validade do polinômio.

        :return: O limite inferior de vazão de jusante
        :rtype: float | None
        """
        return self.data[3]

    @limite_inferior_vazao_jusante.setter
    def limite_inferior_vazao_jusante(self, c: float):
        self.data[3] = c

    @property
    def limite_superior_vazao_jusante(self) -> Optional[float]:
        """
        O limite superior de vazão de jusante (defluência mais lateral)
        para janela de validade do polinômio.

        :return: O limite superior de vazão de jusante
        :rtype: float | None
        """
        return self.data[4]

    @limite_superior_vazao_jusante.setter
    def limite_superior_vazao_jusante(self, c: float):
        self.data[4] = c

    @property
    def coeficiente_a0(self) -> Optional[float]:
        """
        O coeficiente de grau 0 do polinômio.

        :return: O coeficiente de grau 0 do polinômio
        :rtype: float | None
        """
        return self.data[5]

    @coeficiente_a0.setter
    def coeficiente_a0(self, c: float):
        self.data[5] = c

    @property
    def coeficiente_a1(self) -> Optional[float]:
        """
        O coeficiente de grau 1 do polinômio.

        :return: O coeficiente de grau 1 do polinômio
        :rtype: float | None
        """
        return self.data[6]

    @coeficiente_a1.setter
    def coeficiente_a1(self, c: float):
        self.data[6] = c

    @property
    def coeficiente_a2(self) -> Optional[float]:
        """
        O coeficiente de grau 2 do polinômio.

        :return: O coeficiente de grau 2 do polinômio
        :rtype: float | None
        """
        return self.data[7]

    @coeficiente_a2.setter
    def coeficiente_a2(self, c: float):
        self.data[7] = c

    @property
    def coeficiente_a3(self) -> Optional[float]:
        """
        O coeficiente de grau 3 do polinômio.

        :return: O coeficiente de grau 3 do polinômio
        :rtype: float | None
        """
        return self.data[8]

    @coeficiente_a3.setter
    def coeficiente_a3(self, c: float):
        self.data[8] = c

    @property
    def coeficiente_a4(self) -> Optional[float]:
        """
        O coeficiente de grau 4 do polinômio.

        :return: O coeficiente de grau 4 do polinômio
        :rtype: float | None
        """
        return self.data[9]

    @coeficiente_a4.setter
    def coeficiente_a4(self, c: float):
        self.data[9] = c


class HidreletricaCurvaJusanteAfogamentoExplicitoUsina(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-CURVAJUSANTE-AFOGAMENTO-EXPLICITO-USINA"
    IDENTIFIER_DIGITS = 53
    LINE = Line(
        [
            IntegerField(),
            LiteralField(size=3),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina hidrelétrica relacionada ao polinômio.

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def considera_afogamento(self) -> Optional[str]:
        """
        Habilitação do afogamento explícito.

        :return: O flag da habilitação.
        :rtype: str | None
        """
        return self.data[1]

    @considera_afogamento.setter
    def considera_afogamento(self, c: str):
        self.data[1] = c


class HidreletricaCurvaJusanteAfogamentoExplicitoPadrao(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-CURVAJUSANTE-AFOGAMENTO-EXPLICITO-PADRAO"
    IDENTIFIER_DIGITS = 54
    LINE = Line(
        [
            LiteralField(size=3),
        ],
        delimiter=";",
    )

    @property
    def considera_afogamento(self) -> Optional[str]:
        """
        Habilitação do afogamento explícito.

        :return: O flag da habilitação.
        :rtype: str | None
        """
        return self.data[0]

    @considera_afogamento.setter
    def considera_afogamento(self, c: str):
        self.data[0] = c


class HidreletricaProdutibilidadeEspecificaGrade(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-PRODUTIBILIDADE-ESPECIFICA-GRADE"
    IDENTIFIER_DIGITS = 46
    LINE = Line(
        [
            IntegerField(),
            FloatField(size=15, decimal_digits=1),
            FloatField(size=15, decimal_digits=1),
            FloatField(size=15, decimal_digits=6),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def altura_queda_liquida(self) -> Optional[float]:
        """
        A altura de queda líquida relativa à produbilidade.

        :return: A altura (m)
        :rtype: float | None
        """
        return self.data[1]

    @altura_queda_liquida.setter
    def altura_queda_liquida(self, c: float):
        self.data[1] = c

    @property
    def turbinamento(self) -> Optional[float]:
        """
        O turbinamento relativo à produtibilidade.

        :return: O turbinamento (m^3/s)
        :rtype: float | None
        """
        return self.data[2]

    @turbinamento.setter
    def turbinamento(self, c: float):
        self.data[2] = c

    @property
    def produtibilidade_especifica(self) -> Optional[float]:
        """
        A produtibilidade específica da usina.

        :return: A produtibilidade em MW / (m^3/s / m)
        :rtype: float | None
        """
        return self.data[3]

    @produtibilidade_especifica.setter
    def produtibilidade_especifica(self, c: float):
        self.data[3] = c


class HidreletricaPerdaHidraulicaGrade(Register):
    """ """

    IDENTIFIER = "HIDRELETRICA-PERDA-HIDRAULICA-GRADE"
    IDENTIFIER_DIGITS = 36
    LINE = Line(
        [
            IntegerField(),
            FloatField(size=15, decimal_digits=1),
            FloatField(size=15, decimal_digits=1),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def turbinamento(self) -> Optional[float]:
        """
        O turbinamento relativo à produtibilidade.

        :return: O turbinamento (m^3/s)
        :rtype: float | None
        """
        return self.data[1]

    @turbinamento.setter
    def turbinamento(self, c: float):
        self.data[1] = c

    @property
    def perda_hidraulica(self) -> Optional[float]:
        """
        A perda hidráulica em valores absolutos

        :return: A perda (m)
        :rtype: float | None
        """
        return self.data[2]

    @perda_hidraulica.setter
    def perda_hidraulica(self, c: float):
        self.data[2] = c


class EstacaoBombeamento(Register):
    """ """

    IDENTIFIER = "ESTACAO-BOMBEAMENTO"
    IDENTIFIER_DIGITS = 20
    LINE = Line(
        [
            IntegerField(),
            LiteralField(size=30),
            IntegerField(),
            IntegerField(),
            FloatField(size=15, decimal_digits=2),
            FloatField(size=8, decimal_digits=2),
        ],
        delimiter=";",
    )

    @property
    def codigo_estacao(self) -> Optional[int]:
        """
        O código da estação

        :return: O código da estação
        :rtype: int | None
        """
        return self.data[0]

    @codigo_estacao.setter
    def codigo_estacao(self, c: int):
        self.data[0] = c

    @property
    def nome_estacao(self) -> Optional[str]:
        """
        O nome da estação de bomeamento

        :return: O nome como string.
        :rtype: str | None
        """
        return self.data[1]

    @nome_estacao.setter
    def nome_estacao(self, c: str):
        self.data[1] = c

    @property
    def codigo_usina_origem(self) -> Optional[int]:
        """
        O código da usina de origem da estação.

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[2]

    @codigo_usina_origem.setter
    def codigo_usina_origem(self, c: int):
        self.data[2] = c

    @property
    def codigo_usina_destino(self) -> Optional[int]:
        """
        O código da usina de destino da estação.

        :return: O código da usina
        :rtype: int | None
        """
        return self.data[3]

    @codigo_usina_destino.setter
    def codigo_usina_destino(self, c: int):
        self.data[3] = c

    @property
    def consumo_estacao(self) -> Optional[float]:
        """
        O consumo da estação de bomeamento

        :return: O consumo
        :rtype: float | None
        """
        return self.data[4]

    @consumo_estacao.setter
    def consumo_estacao(self, c: float):
        self.data[4] = c

    @property
    def bombeamento_maximo(self) -> Optional[float]:
        """
        O limite de capacidade de bombeamento da estação

        :return: O limite máximo
        :rtype: float | None
        """
        return self.data[5]

    @bombeamento_maximo.setter
    def bombeamento_maximo(self, c: float):
        self.data[5] = c


class EstacaoBombeamentoLimitesPeriodoPatamar(Register):
    """
    Registro que contém os limites de bombeamento de cada
    estação por período e patamar.
    """

    IDENTIFIER = "ESTACAO-BOMBEAMENTO-LIMITES-PERIODO-PATAMAR"
    IDENTIFIER_DIGITS = 44
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            IntegerField(),
            FloatField(size=16, decimal_digits=4, format="e"),
            FloatField(size=16, decimal_digits=4, format="e"),
        ],
        delimiter=";",
    )

    @property
    def codigo_estacao(self) -> Optional[int]:
        """
        O código da estação.

        :return: O código
        :rtype:  int | None
        """
        return self.data[0]

    @codigo_estacao.setter
    def codigo_estacao(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade dos limites da estação.

        :return: A data
        :rtype: datetime | None
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade dos limites da estação.

        :return: A data
        :rtype: datetime | None
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, v: datetime):
        self.data[2] = v

    @property
    def patamar(self) -> Optional[int]:
        """
        O índice do patamar de validade dos limites.

        :return: O patamar
        :rtype: int | None
        """
        return self.data[3]

    @patamar.setter
    def patamar(self, v: int):
        self.data[3] = v

    @property
    def limite_inferior(self) -> Optional[float]:
        """
        O limite inferior para a estação.

        :return: O limite inferior
        :rtype: float | None
        """
        return self.data[4]

    @limite_inferior.setter
    def limite_inferior(self, v: float):
        self.data[4] = v

    @property
    def limite_superior(self) -> Optional[float]:
        """
        O limite superior para a estação.

        :return: O limite superior
        :rtype: float | None
        """
        return self.data[5]

    @limite_superior.setter
    def limite_superior(self, v: float):
        self.data[5] = v


class EstacaoBombeamentoSubmercado(Register):
    """ """

    IDENTIFIER = "ESTACAO-BOMBEAMENTO-SUBMERCADO"
    IDENTIFIER_DIGITS = 31
    LINE = Line(
        [
            IntegerField(),
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def codigo_estacao(self) -> Optional[int]:
        """
        O código da estação

        :return: O código da estação
        :rtype: int | None
        """
        return self.data[0]

    @codigo_estacao.setter
    def codigo_estacao(self, c: int):
        self.data[0] = c

    @property
    def codigo_submercado(self) -> Optional[int]:
        """
        O código do submercado da estação

        :return: O código do submercado
        :rtype: int | None
        """
        return self.data[1]

    @codigo_submercado.setter
    def codigo_submercado(self, c: int):
        self.data[1] = c


class VolumeReferencialTipoPadrao(Register):
    """ """

    IDENTIFIER = "VOLUME-REFERENCIAL-TIPO-PADRAO"
    IDENTIFIER_DIGITS = 30
    LINE = Line(
        [
            IntegerField(),
        ],
        delimiter=";",
    )

    @property
    def tipo_referencia(self) -> Optional[int]:
        """
        O tipo de volume de referência adotado.

        :return: O tipo de volume
        :rtype: int | None
        """
        return self.data[0]

    @tipo_referencia.setter
    def tipo_referencia(self, c: int):
        self.data[0] = c


class VolumeReferencialPeriodo(Register):
    """ """

    IDENTIFIER = "CADH-VOL-REF-PER"
    IDENTIFIER_DIGITS = 16
    LINE = Line(
        [
            IntegerField(),
            DatetimeField(size=7, format="%Y/%m"),
            DatetimeField(size=7, format="%Y/%m"),
            FloatField(size=15, decimal_digits=2),
        ],
        delimiter=";",
    )

    @property
    def codigo_usina(self) -> Optional[int]:
        """
        O código da usina

        :return: O código
        :rtype: int | None
        """
        return self.data[0]

    @codigo_usina.setter
    def codigo_usina(self, c: int):
        self.data[0] = c

    @property
    def data_inicio(self) -> Optional[datetime]:
        """
        A data de início da validade do volume de referência.

        :return: A data
        :rtype: datetime | None
        """
        return self.data[1]

    @data_inicio.setter
    def data_inicio(self, c: datetime):
        self.data[1] = c

    @property
    def data_fim(self) -> Optional[datetime]:
        """
        A data de fim da validade do volume de referência.

        :return: A data
        :rtype: datetime | None
        """
        return self.data[2]

    @data_fim.setter
    def data_fim(self, v: datetime):
        self.data[2] = v

    @property
    def volume_referencia(self) -> Optional[float]:
        """
        O volume de referência da usina.

        :return: O volume
        :rtype: float | None
        """
        return self.data[3]

    @volume_referencia.setter
    def volume_referencia(self, c: float):
        self.data[3] = c



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)

from inewave.nwlistop.modelos.vmortsin import VmortAnos


class Vmortsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes à energia
    de enchimento de volume morto para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vmortsin.out`.

    """

    BLOCKS = [
        VmortAnos,
    ]



from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type


class Abertura(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao número de
    aberturas utilizadas por período.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = []



# Rotinas de testes associadas ao arquivo term.dat do NEWAVE
from inewave.config import MESES_DF
from inewave.newave.modelos.term import BlocoTermUTE

from inewave.newave import Term


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.term import MockBlocoUTE

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_usinas_term():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTE))
    b = BlocoTermUTE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 1625
    assert b.data.iloc[0, 0] == 1
    assert b.data.iloc[-1, -1] == 63.0


def test_atributos_encontrados_term():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTE))
    with patch("builtins.open", m):
        ad = Term.read(ARQ_TESTE)
        assert ad.usinas is not None


def test_atributos_nao_encontrados_term():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Term.read(ARQ_TESTE)
        assert ad.usinas is None


def test_eq_term():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTE))
    with patch("builtins.open", m):
        cf1 = Term.read(ARQ_TESTE)
        cf2 = Term.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_term():
    m: MagicMock = mock_open(read_data="".join(MockBlocoUTE))
    with patch("builtins.open", m):
        cf1 = Term.read(ARQ_TESTE)
        cf2 = Term.read(ARQ_TESTE)
        cf2.usinas.iloc[0, 0] = -1
        assert cf1 != cf2


def test_leitura_escrita_term():
    m_leitura: MagicMock = mock_open(read_data="".join(MockBlocoUTE))
    with patch("builtins.open", m_leitura):
        cf1 = Term.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Term.read(ARQ_TESTE)
        assert cf1 == cf2



# Rotinas de testes associadas ao arquivo MEDIAS-USIT.CSV do NWLISTOP
from inewave.nwlistop.mediasusit import Mediasusit


ARQ_TESTE = "tests/_arquivos/MEDIAS-USIT.CSV"


def test_eq_mediasusit():
    leitor = Mediasusit.read(ARQ_TESTE)
    leitor2 = Mediasusit.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (7680, 5)



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.dlppdfmax import DLPPdfmaxAnos


class Dlppdfmaxm(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes à violação das restrições LPP
    de defluência máxima por patamar, por Submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `dlppdfmax00x.out`, onde x varia conforme o
    Submercado em questão.

    """

    BLOCKS = [
        Submercado,
        DLPPdfmaxAnos,
    ]



from inewave.newave.modelos.sistema import (
    BlocoCustosDeficit,
    BlocoGeracaoUsinasNaoSimuladas,
    BlocoIntercambioSubsistema,
    BlocoMercadoEnergiaSistema,
    BlocoNumeroPatamaresDeficit,
)


from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Sistema(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    dos subsistemas (submercados).

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoNumeroPatamaresDeficit,
        BlocoCustosDeficit,
        BlocoIntercambioSubsistema,
        BlocoMercadoEnergiaSistema,
        BlocoGeracaoUsinasNaoSimuladas,
    ]

    @property
    def numero_patamares_deficit(self) -> Optional[int]:
        """
        O número de patamares de déficit utilizados no estudo.

        :return: O número de patamares como um inteiro
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoNumeroPatamaresDeficit)
        if isinstance(b, BlocoNumeroPatamaresDeficit):
            return b.data
        return None

    @numero_patamares_deficit.setter
    def numero_patamares_deficit(self, n: int):
        b = self.data.get_sections_of_type(BlocoNumeroPatamaresDeficit)
        if isinstance(b, BlocoNumeroPatamaresDeficit):
            b.data = n

    @property
    def custo_deficit(self) -> Optional[pd.DataFrame]:
        """
        Tabela com o custo de cada patamar de déficit, por
        subsistema.

        - codigo_submercado (`int`)
        - nome_submercado (`str`)
        - ficticio (`int`)
        - patamar_deficit (`int`)
        - custo (`float`)
        - corte (`float`)

        :return: A duração por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoCustosDeficit)
        if isinstance(b, BlocoCustosDeficit):
            return b.data
        return None

    @custo_deficit.setter
    def custo_deficit(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoCustosDeficit)
        if isinstance(b, BlocoCustosDeficit):
            b.data = df

    @property
    def limites_intercambio(self) -> Optional[pd.DataFrame]:
        """
        Tabela com o limite de intercâmbio por par de
        subsistemas.

        - submercado_de (`int`)
        - submercado_para (`int`)
        - sentido (`int`)
        - data (`datetime`)
        - valor (`float`)

        :return: A duração por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoIntercambioSubsistema)
        if isinstance(b, BlocoIntercambioSubsistema):
            return b.data
        return None

    @limites_intercambio.setter
    def limites_intercambio(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoIntercambioSubsistema)
        if isinstance(b, BlocoIntercambioSubsistema):
            b.data = df

    @property
    def mercado_energia(self) -> Optional[pd.DataFrame]:
        """
        Tabela com o mercado total de energia por período de estudo.

        - codigo_submercado (`int`)
        - data (`datetime`)
        - valor (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoMercadoEnergiaSistema)
        if isinstance(b, BlocoMercadoEnergiaSistema):
            return b.data
        return None

    @mercado_energia.setter
    def mercado_energia(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoMercadoEnergiaSistema)
        if isinstance(b, BlocoMercadoEnergiaSistema):
            b.data = df

    @property
    def geracao_usinas_nao_simuladas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a geração das usinas não simuladas por fonte
        de geração.

        - codigo_submercado (`int`)
        - indice_bloco (`int`)
        - fonte (`str`)
        - data (`int`)
        - valor (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoGeracaoUsinasNaoSimuladas)
        if isinstance(b, BlocoGeracaoUsinasNaoSimuladas):
            return b.data
        return None

    @geracao_usinas_nao_simuladas.setter
    def geracao_usinas_nao_simuladas(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoGeracaoUsinasNaoSimuladas)
        if isinstance(b, BlocoGeracaoUsinasNaoSimuladas):
            b.data = df



import pandas as pd  # type: ignore

from cfinterface.files.sectionfile import SectionFile

from inewave.nwlistop.modelos.mediasusie import TabelaMediasusie


class Mediasusie(SectionFile):
    """
    Armazena os dados das saídas referentes às médias de diversas variáveis
    de cada estação elevatória.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `MEDIAS-USIE.CSV`.

    """

    SECTIONS = [TabelaMediasusie]

    @property
    def valores(self) -> pd.DataFrame:
        """
        Tabela com os valores de médias para as variáveis das estações
        elevatórias.

        - estagio (`int`)
        - codigo_usina (`int`)
        - <variavel_1> (`float`)
        - ...
        - <variavel_n> (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(TabelaMediasusie)
        if isinstance(b, TabelaMediasusie):
            return b.data
        return None



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from typing import List, IO


class BlocoDadosSelcor(Section):
    """
    Bloco de informações da parametrização do processo de seleção
    de cortes existentes no arquivo `selcor.dat`.
    """

    __slots__ = ["__linha", "__cabecalhos", "__comentarios", "data"]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                LiteralField(60, 0),
                IntegerField(4, 63),
                IntegerField(4, 69),
            ]
        )
        self.__cabecalhos: List[str] = []
        self.__comentarios: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoDadosSelcor):
            return False
        bloco: BlocoDadosSelcor = o
        if not all(
            [
                isinstance(self.data, list),
                isinstance(o.data, list),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    # Override
    def read(self, file: IO, *args, **kwargs):
        # Salta as linhas de cabeçalhos
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        # Lê as linhas de parâmetros
        self.data: List[List[int]] = []
        for _ in range(7):
            dados_linha = self.__linha.read(file.readline())
            self.__comentarios.append(dados_linha[0])
            self.data.append(dados_linha[1:])

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, list):
            raise ValueError("Dados do selcor.dat não foram lidos com sucesso")

        for c, s in zip(self.__comentarios, self.data):
            file.write(self.__linha.write([c] + s))



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (  # noqa
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.cviol_eletricasin import CviolEletricasinAnos


class CviolEletricasin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes ao custo de violação das
    restrições elétricas para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `cviol_eletricasin.out`.

    """

    BLOCKS = [
        CviolEletricasinAnos,
    ]



# Rotinas de testes associadas ao arquivo MEDIAS-MERC.CSV do NWLISTOP
from inewave.nwlistop.mediasmerc import Mediasmerc
import pandas as pd  # type: ignore


ARQ_TESTE = "tests/_arquivos/MEDIAS-MERC.CSV"


def test_eq_mediasmerc():
    leitor = Mediasmerc.read(ARQ_TESTE)
    leitor2 = Mediasmerc.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (228, 54)



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.defsin import DefAnos


class Defsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes ao déficit
    por patamar para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `defsinp001.out`.
    """

    BLOCKS = [
        DefAnos,
    ]



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)
from inewave.nwlistop.modelos.verturbm import VertAnos


class Verturbm(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes às energias
    vertidas, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vertub00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        Submercado,
        VertAnos,
    ]



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from typing import List, Dict, IO, Any


class BlocoDadosNwlistop(Section):
    """
    Bloco do arquivo nwlistop.dat que armazena as informações de mês de
    início e fim para a impressão.
    """

    __slots__ = [
        "__linha_opcao",
        "__linha_arquivo",
        "__linha_series_op1",
        "__linha_periodos",
        "__linha_variaveis_op2",
        "__linha_uhes_op2",
        "data",
        "__comentarios",
    ]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_opcao = Line([IntegerField(1, 1)])
        self.__linha_arquivo = Line(
            [LiteralField(29, 0), LiteralField(40, 30)]
        )
        self.__linha_series_op1 = Line(
            [IntegerField(4, 1), IntegerField(4, 6)]
        )
        self.__linha_periodos = Line([IntegerField(3, 1), IntegerField(3, 5)])
        self.__linha_variaveis_op2 = Line(
            [IntegerField(2, 1 + 3 * i) for i in range(21)]
        )
        self.__linha_uhes_op2 = Line(
            [IntegerField(3, 1 + 4 * i) for i in range(16)]
        )
        self.data: Dict[str, Any] = {}
        self.__comentarios: List[List[str]] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoDadosNwlistop):
            return False
        bloco: BlocoDadosNwlistop = o
        if not all(
            [
                isinstance(self.data, dict),
                isinstance(o.data, dict),
            ]
        ):
            return False
        else:
            return self.data == bloco.data

    def __read_op1(self, file: IO):
        # Le os nomes dos arquivos
        self.data["arquivos"] = []
        for _ in range(3):
            self.data["arquivos"].append(
                self.__linha_arquivo.read(file.readline())
            )

        self.__comentarios.append([])
        for _ in range(2):
            self.__comentarios[-1].append(file.readline())
        # Le o periodo inicial e final
        self.data["periodos"] = self.__linha_periodos.read(file.readline())

        self.__comentarios.append([])
        self.__comentarios[-1].append(file.readline())
        # Le a serie inicial e final
        self.data["series"] = self.__linha_series_op1.read(file.readline())

    def __read_op2(self, file: IO):
        # Le os nomes dos arquivos
        self.data["arquivos"] = []
        for _ in range(3):
            self.data["arquivos"].append(
                self.__linha_arquivo.read(file.readline())
            )

        self.__comentarios.append([])
        for _ in range(2):
            self.__comentarios[-1].append(file.readline())
        # Le o periodo inicial e final
        self.data["periodos"] = self.__linha_periodos.read(file.readline())

        self.__comentarios.append([])
        for _ in range(6):
            self.__comentarios[-1].append(file.readline())
        # Le as variáveis para os estágios agregados
        self.data["variaveis_ree"] = self.__linha_variaveis_op2.read(
            file.readline()
        )

        self.__comentarios.append([])
        for _ in range(4):
            self.__comentarios[-1].append(file.readline())

        # Le as variáveis para os estágios individualizados
        self.data["variaveis_uhe"] = self.__linha_variaveis_op2.read(
            file.readline()
        )

        self.__comentarios.append([])
        self.__comentarios[-1].append(file.readline())

        # Le as usinas para os estágios individualizados
        self.data["uhes"] = self.__linha_uhes_op2.read(file.readline())

    def __read_op4(self, file: IO):
        # Le os nomes dos arquivos
        self.data["arquivos"] = []
        for _ in range(3):
            self.data["arquivos"].append(
                self.__linha_arquivo.read(file.readline())
            )

        self.__comentarios = []
        while True:
            linha = file.readline()
            if len(linha) < 3:
                break
            self.__comentarios.append(linha)

    # Override
    def read(self, file: IO, *args, **kwargs):
        opcao = self.__linha_opcao.read(file.readline())[0]
        self.data["opcao"] = opcao
        if opcao == 1:
            self.__read_op1(file)
        elif opcao == 2:
            self.__read_op2(file)
        elif opcao == 4:
            self.__read_op4(file)

    # Override
    def write(self, file: IO, *args, **kwargs):
        opcao = self.data["opcao"]
        file.write(self.__linha_opcao.write([opcao]))
        if opcao == 1:
            self.__write_op1(file)
        elif opcao == 2:
            self.__write_op2(file)
        elif opcao == 4:
            self.__write_op4(file)

    # Override
    def __write_op1(self, file: IO, *args, **kwargs):
        # Escreve os nomes de arquivos
        for c in self.data["arquivos"]:
            file.write(self.__linha_arquivo.write(c))
        # Escreve as linhas de cabeçalho para períodos
        for c in self.__comentarios[0]:
            file.write(c)
        # Escreve os períodos
        file.write(self.__linha_periodos.write(self.data["periodos"]))
        # Escreve as linhas de cabeçalho para séries
        for c in self.__comentarios[1]:
            file.write(c)
        # Escreve as séries
        file.write(self.__linha_periodos.write(self.data["series"]))

    # Override
    def __write_op2(self, file: IO, *args, **kwargs):
        # Escreve os nomes de arquivos
        for c in self.data["arquivos"]:
            file.write(self.__linha_arquivo.write(c))
        # Escreve as linhas de cabeçalho para períodos
        for c in self.__comentarios[0]:
            file.write(c)
        # Escreve os períodos
        file.write(self.__linha_periodos.write(self.data["periodos"]))
        # Escreve as linhas de cabeçalho para variáveis agregadas
        for c in self.__comentarios[1]:
            file.write(c)
        # Escreve as séries
        file.write(self.__linha_periodos.write(self.data["variaveis_ree"]))
        # Escreve as linhas de cabeçalho para variáveis individualizadas
        for c in self.__comentarios[2]:
            file.write(c)
        # Escreve as séries
        file.write(self.__linha_periodos.write(self.data["variaveis_uhe"]))
        # Escreve as linhas de cabeçalho para usinas
        for c in self.__comentarios[3]:
            file.write(c)
        # Escreve as séries
        file.write(self.__linha_periodos.write(self.data["uhes"]))

    # Override
    def __write_op4(self, file: IO, *args, **kwargs):
        # Escreve os nomes de arquivos
        for c in self.data["arquivos"]:
            file.write(self.__linha_arquivo.write(c))
        # Escreve as linhas de comentários
        for c in self.__comentarios:
            file.write(c)



from inewave.newave.cortes import Cortes

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch
import pytest
import pandas as pd

ARQ_TESTE_REE = "./tests/mocks/arquivos/cortes_ree.dat"
ARQ_TESTE_HIB = "./tests/mocks/arquivos/cortes_hib.dat"

TAMANHO_CORTE_REE = 1664
REES_CORTE = [1, 6, 7, 5, 10, 12, 2, 11, 3, 4, 8, 9]
UHES_CORTE = list(range(154))
SUBMERCADOS_CORTE = [1, 2, 3, 4]
TAMANHO_CORTE_HIB = 17568


def test_atributos_encontrados_cortes_ree():
    h = Cortes.read(
        ARQ_TESTE_REE,
        tamanho_registro=TAMANHO_CORTE_REE,
        indice_ultimo_corte=1,
        numero_total_cortes=1,
        codigos_rees=REES_CORTE,
        codigos_submercados=SUBMERCADOS_CORTE,
    )
    assert isinstance(h.cortes, pd.DataFrame)


def test_atributos_encontrados_cortes_hib():
    h = Cortes.read(
        ARQ_TESTE_HIB,
        tamanho_registro=TAMANHO_CORTE_HIB,
        indice_ultimo_corte=1,
        numero_total_cortes=1,
        codigos_uhes=UHES_CORTE,
        codigos_submercados=SUBMERCADOS_CORTE,
    )
    assert isinstance(h.cortes, pd.DataFrame)


def test_eq_cortes():
    h1 = Cortes.read(
        ARQ_TESTE_HIB,
        tamanho_registro=TAMANHO_CORTE_HIB,
        indice_ultimo_corte=1,
        numero_total_cortes=1,
        codigos_uhes=UHES_CORTE,
        codigos_submercados=SUBMERCADOS_CORTE,
    )
    h2 = Cortes.read(
        ARQ_TESTE_HIB,
        tamanho_registro=TAMANHO_CORTE_HIB,
        indice_ultimo_corte=1,
        numero_total_cortes=1,
        codigos_uhes=UHES_CORTE,
        codigos_submercados=SUBMERCADOS_CORTE,
    )
    assert h1 == h2


def test_atributos_cortes_ree():
    h = Cortes.read(
        ARQ_TESTE_REE,
        tamanho_registro=TAMANHO_CORTE_REE,
        indice_ultimo_corte=1,
        numero_total_cortes=1,
        codigos_rees=REES_CORTE,
        codigos_submercados=SUBMERCADOS_CORTE,
    )
    assert h.cortes.at[0, "indice_corte"] == 1
    assert h.cortes.at[0, "iteracao_construcao"] == 1
    assert h.cortes.at[0, "indice_forward"] == 1
    assert h.cortes.at[0, "iteracao_desativacao"] == 0
    assert h.cortes.at[0, "rhs"] == 1017842.32745
    assert h.cortes.at[0, "pi_earm_ree1"] == -0.00015
    assert h.cortes.at[0, "pi_earm_ree12"] == -0.00021
    assert h.cortes.at[0, "pi_ena_ree1_lag1"] == 0.00084
    assert h.cortes.at[0, "pi_ena_ree7_lag1"] == -0.00013
    assert h.cortes.at[0, "pi_gnl_sbm1_pat1_lag1"] == -0.00021
    assert h.cortes.at[0, "pi_gnl_sbm2_pat3_lag1"] == -0.00191


def test_atributos_cortes_hib():
    h = Cortes.read(
        ARQ_TESTE_HIB,
        tamanho_registro=TAMANHO_CORTE_HIB,
        indice_ultimo_corte=1,
        numero_total_cortes=1,
        codigos_uhes=UHES_CORTE,
        codigos_submercados=SUBMERCADOS_CORTE,
    )
    assert h.cortes.at[0, "indice_corte"] == 1
    assert h.cortes.at[0, "iteracao_construcao"] == 100
    assert h.cortes.at[0, "indice_forward"] == 200
    assert h.cortes.at[0, "iteracao_desativacao"] == 100
    assert h.cortes.at[0, "rhs"] == 279953222.04742
    assert h.cortes.at[0, "pi_gnl_sbm1_pat1_lag1"] == -1.21847
    assert h.cortes.at[0, "pi_gnl_sbm3_pat3_lag2"] == -2.46932
    assert h.cortes.at[0, "pi_varm_uhe0"] == -6.74723
    assert h.cortes.at[0, "pi_varm_uhe153"] == -0.37228
    assert h.cortes.at[0, "pi_qafl_uhe0_lag12"] == 0.53852



from inewave.nwlistop.modelos.blocos.parsubmercados import ParSubmercados
from inewave.nwlistop.modelos.arquivos.arquivoparsubmercadopatamar import (
    ArquivoParSubmercadoPatamar,
)
from inewave.nwlistop.modelos.intercambio import IntercambioAnos


class Intercambio(ArquivoParSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes ao intercâmbio
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `int00x00y.out`, onde x e y variam
    conforme os submercados em questão.

    """

    BLOCKS = [
        ParSubmercados,
        IntercambioAnos,
    ]



from inewave.newave.modelos.vazinat import SecaoDadosVazinat
from inewave.newave.vazinat import Vazinat
from inewave.config import MAX_ANOS_HISTORICO

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch


ARQ_TESTE = "./tests/mocks/arquivos/vazinat.dat"

NUM_CONFIGURACOES = 1
NUM_UHES = 164
ANO_INICIO_HISTORICO = 1931
NUM_ENTRADAS = NUM_CONFIGURACOES * NUM_UHES * 12 * MAX_ANOS_HISTORICO


def test_secao_vazinat():
    r = SecaoDadosVazinat()
    with open(ARQ_TESTE, "rb") as fp:
        r.read(
            fp,
            numero_uhes=NUM_UHES,
            ano_inicio_historico=ANO_INICIO_HISTORICO,
            numero_configuracoes=NUM_CONFIGURACOES,
        )

    assert len(r.data) == NUM_ENTRADAS


def test_atributos_encontrados_vazinat():
    h = Vazinat.read(
        ARQ_TESTE,
        numero_uhes=NUM_UHES,
        ano_inicio_historico=ANO_INICIO_HISTORICO,
        numero_configuracoes=NUM_CONFIGURACOES,
    )
    assert h.series is not None
    assert h.series.isna().sum().sum() == 0


def test_atributos_nao_encontrados_vazinat():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        h = Vazinat.read(
            ARQ_TESTE,
            numero_uhes=NUM_UHES,
            ano_inicio_historico=ANO_INICIO_HISTORICO,
            numero_configuracoes=NUM_CONFIGURACOES,
        )
        assert h.series.isna().sum().sum() == NUM_ENTRADAS


def test_eq_vazinat():
    h1 = Vazinat.read(
        ARQ_TESTE,
        numero_uhes=NUM_UHES,
        ano_inicio_historico=ANO_INICIO_HISTORICO,
        numero_configuracoes=NUM_CONFIGURACOES,
    )
    h2 = Vazinat.read(
        ARQ_TESTE,
        numero_uhes=NUM_UHES,
        ano_inicio_historico=ANO_INICIO_HISTORICO,
        numero_configuracoes=NUM_CONFIGURACOES,
    )
    assert h1 == h2



# Rotinas de testes associadas ao arquivo MEDIAS-SIN.CSV do NWLISTOP
from inewave.nwlistop.mediassin import Mediassin


ARQ_TESTE = "tests/_arquivos/MEDIAS-SIN.CSV"


def test_eq_mediassin():
    leitor = Mediassin.read(ARQ_TESTE)
    leitor2 = Mediassin.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (57, 55)



from typing import Type, TypeVar, List, Optional, Union
from datetime import datetime
import pandas as pd  # type: ignore
from cfinterface.components.register import Register
from cfinterface.files.registerfile import RegisterFile
from inewave.libs.modelos.restricoes import (
    RegistroRE,
    RegistroREHorizPer,
    RegistroRELimFormPer,
    RegistroRHE,
    RegistroRHEHorizPer,
    RegistroRHELsLPPEarmi,
    RegistroRHQ,
    RegistroRHQHorizPer,
    RegistroRHQLsLPPVoli,
    RegistroRHQLimFormPerPat,
    RegistroRHV,
    RegistroRHVHorizPer,
    RegistroRHVLimFormPer,
)


class Restricoes(RegisterFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos dados
    das restrições do problema.
    """

    T = TypeVar("T", bound=Register)

    REGISTERS = [
        RegistroRELimFormPer,
        RegistroREHorizPer,
        RegistroRE,
        RegistroRHELsLPPEarmi,
        RegistroRHEHorizPer,
        RegistroRHE,
        RegistroRHQLimFormPerPat,
        RegistroRHQLsLPPVoli,
        RegistroRHQHorizPer,
        RegistroRHQ,
        RegistroRHVLimFormPer,
        RegistroRHVHorizPer,
        RegistroRHV,
    ]

    def __registros_ou_df(
        self, t: Type[T], **kwargs
    ) -> Optional[Union[T, List[T], pd.DataFrame]]:
        if kwargs.get("df"):
            return self._as_df(t)
        else:
            kwargs_sem_df = {k: v for k, v in kwargs.items() if k != "df"}
            return self.data.get_registers_of_type(t, **kwargs_sem_df)

    def re(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[RegistroRE, List[RegistroRE], pd.DataFrame]]:
        """
        Obtém um registro que cadastra uma usina restrição elétrica (RE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRE` |
            list[:class:`RegistroRE`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRE,
            codigo_restricao=codigo_restricao,
            formula=formula,
            df=df,
        )

    def re_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroREHorizPer, List[RegistroREHorizPer], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição elétrica.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data inicial de validade da restrição
        :type data_inicio: datetime | None
        :param data_fim: data final de validade da restrição
        :type data_fim: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroREHorizPer` |
            list[:class:`RegistroREHorizPer`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroREHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )

    def re_lim_form_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        patamar: Optional[int] = None,
        limite_inferior: Optional[float] = None,
        limite_superior: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroRELimFormPer, List[RegistroRELimFormPer], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra os limites por horizonte e por
        patamar para uma restrição elétrica.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data inicial de validade dos limites
        :type data_inicio: datetime | None
        :param data_fim: data final de validade dos limites
        :type data_fim: datetime | None
        :param patamar: patamar de validade dos limites
        :type patamar: int | None
        :param limite_inferior: limite inferior da restrição
        :type limite_inferior: float | None
        :param limite_superior: limite superior da restrição
        :type limite_superior: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRELimFormPer` |
            list[:class:`RegistroRELimFormPer`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRELimFormPer,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            patamar=patamar,
            limite_inferior=limite_inferior,
            limite_superior=limite_superior,
            df=df,
        )

    def rhe(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[RegistroRHE, List[RegistroRHE], pd.DataFrame]]:
        """
        Obtém um registro que cadastra uma usina restrição linear por
        partes de energia (REE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHE` |
            list[:class:`RegistroRHE`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHE,
            codigo_restricao=codigo_restricao,
            formula=formula,
            df=df,
        )

    def rhe_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroRHEHorizPer, List[RegistroRHEHorizPer], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição linear por partes de energia.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data inicial de validade da restrição
        :type data_inicio: datetime | None
        :param data_fim: data final de validade da restrição
        :type data_fim: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHEHorizPer` |
            list[:class:`RegistroRHEHorizPer`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHEHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )

    def rhe_ls_lpp_earmi(
        self,
        codigo_restricao: Optional[int] = None,
        indice_reta: Optional[int] = None,
        coeficiente_angular: Optional[float] = None,
        coeficiente_linear: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroRHELsLPPEarmi, List[RegistroRHELsLPPEarmi], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra as retas da restrição linear
        por partes de energia.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param indice_reta: índice da reta definida
        :type indice_reta: int | None
        :param coeficiente_angular: coeficiente angular da reta
        :type coeficiente_angular: float | None
        :param coeficiente_linear: coeficiente linear da reta
        :type coeficiente_linear: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHELsLPPEarmi` |
            list[:class:`RegistroRHELsLPPEarmi`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHELsLPPEarmi,
            codigo_restricao=codigo_restricao,
            indice_reta=indice_reta,
            coeficiente_angular=coeficiente_angular,
            coeficiente_linear=coeficiente_linear,
            df=df,
        )

    def rhq(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[RegistroRHQ, List[RegistroRHQ], pd.DataFrame]]:
        """
        Obtém um registro que cadastra uma usina restrição de vazão (UHE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQ` |
            list[:class:`RegistroRHQ`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHQ,
            codigo_restricao=codigo_restricao,
            formula=formula,
            df=df,
        )

    def rhq_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroRHQHorizPer, List[RegistroRHQHorizPer], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição de vazão.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data inicial de validade da restrição
        :type data_inicio: datetime | None
        :param data_fim: data final de validade da restrição
        :type data_fim: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQHorizPer` |
            list[:class:`RegistroRHQHorizPer`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHQHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )

    def rhq_ls_lpp_voli(
        self,
        codigo_restricao: Optional[int] = None,
        indice_reta: Optional[int] = None,
        coeficiente_angular: Optional[float] = None,
        coeficiente_linear: Optional[float] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroRHQLsLPPVoli, List[RegistroRHQLsLPPVoli], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra as retas da restrição linear
        por partes de vazão.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param indice_reta: índice da reta definida
        :type indice_reta: int | None
        :param coeficiente_angular: coeficiente angular da reta
        :type coeficiente_angular: float | None
        :param coeficiente_linear: coeficiente linear da reta
        :type coeficiente_linear: float | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQLsLPPVoli` |
            list[:class:`RegistroRHQLsLPPVoli`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHQLsLPPVoli,
            codigo_restricao=codigo_restricao,
            indice_reta=indice_reta,
            coeficiente_angular=coeficiente_angular,
            coeficiente_linear=coeficiente_linear,
            df=df,
        )

    def rhq_lim_form_per_pat(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        patamar: Optional[int] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroRHQLimFormPerPat,
            List[RegistroRHQLimFormPerPat],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que cadastra as retas da restrição de vazão.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data de início da validade dos limites
        :type data_inicio: datetime | None
        :param data_fim: data de fim da validade dos limites
        :type data_fim: datetime | None
        :param patamar: patamar de validade dos limites
        :type patamar: int | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHQLimFormPerPat` |
            list[:class:`RegistroRHQLimFormPerPat`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHQLimFormPerPat,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            patamar=patamar,
            df=df,
        )

    def rhv(
        self,
        codigo_restricao: Optional[int] = None,
        formula: Optional[str] = None,
        df: bool = False,
    ) -> Optional[Union[RegistroRHV, List[RegistroRHV], pd.DataFrame]]:
        """
        Obtém um registro que cadastra uma usina restrição de volume (UHE).

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param formula: equação da restrição
        :type formula: str | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHV` |
            list[:class:`RegistroRHV`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHV,
            codigo_restricao=codigo_restricao,
            formula=formula,
            df=df,
        )

    def rhv_horiz_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[RegistroRHVHorizPer, List[RegistroRHVHorizPer], pd.DataFrame]
    ]:
        """
        Obtém um registro que cadastra o horizonte de validade de uma
        restrição de volume.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data inicial de validade da restrição
        :type data_inicio: datetime | None
        :param data_fim: data final de validade da restrição
        :type data_fim: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHVHorizPer` |
            list[:class:`RegistroRHVHorizPer`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHVHorizPer,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )

    def rhv_lim_form_per(
        self,
        codigo_restricao: Optional[int] = None,
        data_inicio: Optional[datetime] = None,
        data_fim: Optional[datetime] = None,
        df: bool = False,
    ) -> Optional[
        Union[
            RegistroRHVLimFormPer,
            List[RegistroRHVLimFormPer],
            pd.DataFrame,
        ]
    ]:
        """
        Obtém um registro que cadastra as retas da restrição de volume.

        :param codigo_restricao: código que especifica a restrição
        :type codigo_restricao: int | None
        :param data_inicio: data de início da validade dos limites
        :type data_inicio: datetime | None
        :param data_fim: data de fim da validade dos limites
        :type data_fim: datetime | None
        :return: Um ou mais registros, se existirem.
        :rtype: :class:`RegistroRHVLimFormPer` |
            list[:class:`RegistroRHVLimFormPer`] | `pd.DataFrame` | None
        """
        return self.__registros_ou_df(
            RegistroRHVLimFormPer,
            codigo_restricao=codigo_restricao,
            data_inicio=data_inicio,
            data_fim=data_fim,
            df=df,
        )



from inewave.nwlistop.modelos.blocos.estacaobombeamento import (
    EstacaoBombeamento,
)
from inewave.nwlistop.modelos.arquivos.arquivoestacaobombeamentopatamar import (  # noqa
    ArquivoEstacaoBombeamentoPatamar,
)
from inewave.nwlistop.modelos.qbomb import QBombAnos


class Qbomb(ArquivoEstacaoBombeamentoPatamar):
    """
    Armazena os dados das saídas referentes à vazão bombeada
    por estação de bombeamento.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `qbombXXX.out`.

    """

    BLOCKS = [
        EstacaoBombeamento,
        QBombAnos,
    ]



from inewave.newave.modelos.re import (
    BlocoUsinasConjuntoRE,
    BlocoConfiguracaoRestricoesRE,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Re(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às restrições
    elétricas existentes.

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoUsinasConjuntoRE,
        BlocoConfiguracaoRestricoesRE,
    ]

    @property
    def usinas_conjuntos(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os conjuntos de usinas com restrições elétricas.

        - conjunto (`int`)
        - codigo_usina (`float`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUsinasConjuntoRE)
        if isinstance(b, BlocoUsinasConjuntoRE):
            return b.data
        return None

    @usinas_conjuntos.setter
    def usinas_conjuntos(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUsinasConjuntoRE)
        if isinstance(b, BlocoUsinasConjuntoRE):
            b.data = df

    @property
    def restricoes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as configurações das restrições elétricas.

        - conjunto (`int`)
        - mes_inicio (`int`)
        - mes_fim (`int`)
        - ano_inicio (`int`)
        - ano_fim (`int`)
        - patamar (`int`)
        - restricao (`float`)
        - motivo (`str`)

        :return: A tabela como um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoConfiguracaoRestricoesRE)
        if isinstance(b, BlocoConfiguracaoRestricoesRE):
            return b.data
        return None

    @restricoes.setter
    def restricoes(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoConfiguracaoRestricoesRE)
        if isinstance(b, BlocoConfiguracaoRestricoesRE):
            b.data = df



from inewave.newave.nwv_eco_evap import NwvEcoEvap

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.nwv_eco_evap import MockNwvEcoEvap

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_nwv_eco_evap():
    m: MagicMock = mock_open(read_data="".join(MockNwvEcoEvap))
    with patch("builtins.open", m):
        rel = NwvEcoEvap.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "periodo"] == 1
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "volume_referencia_hm3"] == 265.86
        assert rel.tabela.at[0, "evaporacao_referencia_hm3"] == 0.23
        assert rel.tabela.at[0, "coeficiente_evaporacao_mm_mes"] == 6
        assert rel.tabela.at[0, "flag_evaporacao"] == 1
        assert rel.tabela.at[0, "evaporacao_linear"] == 1
        assert rel.tabela.at[0, "tipo_volume_referencia"] == 1


def test_eq_nwv_eco_evap():
    m: MagicMock = mock_open(read_data="".join(MockNwvEcoEvap))
    with patch("builtins.open", m):
        rel1 = NwvEcoEvap.read(ARQ_TESTE)
        rel2 = NwvEcoEvap.read(ARQ_TESTE)
        assert rel1 == rel2


def test_neq_nwv_eco_evap():
    m: MagicMock = mock_open(read_data="".join(MockNwvEcoEvap))
    with patch("builtins.open", m):
        rel1 = NwvEcoEvap.read(ARQ_TESTE)
        rel2 = NwvEcoEvap.read(ARQ_TESTE)
        rel1.tabela.iloc[0, 0] = -1
        assert rel1 != rel2



from inewave.newave.modelos.term import BlocoTermUTE

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Term(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos dados das
    usinas térmicas.
    """

    T = TypeVar("T")

    SECTIONS = [BlocoTermUTE]

    @property
    def usinas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com configurações e inflexibilidades das usinas térmicas.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - potencia_instalada (`float`)
        - fator_capacidade_maximo (`float`)
        - teif (`float`)
        - indisponibilidade_programada (`float`)
        - mes (`int`)
        - geracao_minima (`float`)

        **OBS:** O mês de validade para a geração mínima
        pertence ao intervalo [1, 13], onde 13 vale para
        o 13º mês em diante.

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoTermUTE)
        if isinstance(b, BlocoTermUTE):
            return b.data
        return None

    @usinas.setter
    def usinas(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoTermUTE)
        if isinstance(b, BlocoTermUTE):
            b.data = df



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (  # noqa
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.viol_eletricasin import ViolEeletricasinAnos


class ViolEletricasin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação das
    restrições elétricas para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `viol_eletricasin.out`.

    """

    BLOCKS = [
        ViolEeletricasinAnos,
    ]



from inewave.newave.evap_cortes import EvapCortes

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.nwv_cortes_evap import MockNwvCortesEvap

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_evap_cortes():
    m: MagicMock = mock_open(read_data="".join(MockNwvCortesEvap))
    with patch("builtins.open", m):
        rel = EvapCortes.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "periodo"] == 1
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "derivada_cota_area"] == 0.0000000000
        assert rel.tabela.at[0, "derivada_volume_cota"] == 0.0000000000
        assert rel.tabela.at[0, "volume_referencia_hm3"] == 265.86
        assert rel.tabela.at[0, "evaporacao_referencia_hm3"] == 0.23
        assert rel.tabela.at[0, "coeficiente_volume"] == 0.0000000000
        assert rel.tabela.at[0, "rhs_volume"] == 0.2263


def test_eq_evap_cortes():
    m: MagicMock = mock_open(read_data="".join(MockNwvCortesEvap))
    with patch("builtins.open", m):
        rel1 = EvapCortes.read(ARQ_TESTE)
        rel2 = EvapCortes.read(ARQ_TESTE)
        assert rel1 == rel2


def test_neq_evap_cortes():
    m: MagicMock = mock_open(read_data="".join(MockNwvCortesEvap))
    with patch("builtins.open", m):
        rel1 = EvapCortes.read(ARQ_TESTE)
        rel2 = EvapCortes.read(ARQ_TESTE)
        rel1.tabela.iloc[0, 0] = -1
        assert rel1 != rel2



from inewave.newave.modelos.blocos.versaomodelo import VersaoModelo
from inewave.newave.modelos.eco_fpha import TabelaEcoFpha

from inewave.newave.modelos.arquivoscsv.arquivocsv import ArquivoCSV
from typing import Optional
import pandas as pd  # type: ignore


class FphaEco(ArquivoCSV):
    """
    Arquivo com o eco da função de produção para as UHEs
    do NEWAVE.
    """

    BLOCKS = [VersaoModelo, TabelaEcoFpha]

    @property
    def tabela(self) -> Optional[pd.DataFrame]:
        """
        A tabela de dados que está contida no arquivo.

        - codigo_usina (`int`)
        - periodo (`int`)
        - nome_usina (`str`)
        - tipo (`int`)
        - conv (`int`)
        - alfa (`int`)
        - rems (`int`)
        - numero_pontos_vazao_turbinada (`int`)
        - vazao_turbinada_minima (`float`)
        - vazao_turbinada_maxima (`float`)
        - numero_pontos_volume_armazenado (`int`)
        - volume_armazenado_minimo (`float`)
        - volume_armazenado_maximo (`float`)
        - geracao_minima (`float`)
        - geracao_maxima (`float`)

        :return: A tabela como um dataframe
        :rtype: pd.DataFrame | None
        """
        return self._tabela()



from inewave.newave.modelos.blocos.versaomodelo import (
    VersaoModelo,
    VersaoModeloLibs,
)
from inewave.newave.modelos.blocos.tabelacsv import TabelaCSV

from cfinterface.components.block import Block
from cfinterface.files.blockfile import BlockFile
import pandas as pd  # type: ignore
from typing import List, Type, TypeVar, Optional


class ArquivoCSV(BlockFile):
    """
    Modelo de arquivo baseado em blocos específico para o formato
    dos arquivos de saída .CSV do NEWAVE. Espera conter
    a implementação de dois blocos específicos:

    - Versão do modelo
    - Tabela de dados

    O primeiro é genérico, mas o segundo deve ser
    implementado para cada arquivo específico a ser lido.
    """

    BLOCKS: List[Type[Block]] = [VersaoModelo]
    ENCODING = "iso-8859-1"

    T = TypeVar("T")

    @property
    def versao(self) -> Optional[str]:
        """
        A versão do modelo utilizada para executar o caso.

        :return: A versão do modelo
        :rtype: str | None
        """
        b = self.data.get_blocks_of_type(VersaoModelo)
        b_libs = self.data.get_blocks_of_type(VersaoModeloLibs)
        if isinstance(b, VersaoModelo):
            return b.data
        elif isinstance(b_libs, VersaoModeloLibs):
            return b_libs.data
        return None

    def _tabela(self) -> Optional[pd.DataFrame]:
        """
        A tabela de dados que está contida no arquivo.

        :return: A tabela como um dataframe
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(TabelaCSV)
        if isinstance(b, TabelaCSV):
            return b.data
        return None



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.exces import ExcesAnos


class Fteolsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à folga da variável de
    geração eólica para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `fteolsin.out`

    """

    BLOCKS = [
        ExcesAnos,
    ]



from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)
from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.cmargmed import CmargsAnos, CmargsAnos28


class Cmargmed(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes aos custos marginais de operação
    por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cmarg00x-med.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        CmargsAnos,
    ]

    VERSIONS = {
        "28": [
            Submercado,
            CmargsAnos28,
        ],
        "29.4.1": [Submercado, CmargsAnos],
    }



from inewave.nwlistop.modelos.custo_futuro import CustoFuturoAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class CustoFuturo(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao custo futuro
    de cada estágio em cada série.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `custo_futuro.out`.
    """

    BLOCKS = [
        CustoFuturoAnos,
    ]



MockArquivos = [
    "ARQUIVO DE DADOS GERAIS     : nwlistcf.dat\n",
    "ARQUIVO DE CORTES DE BENDERS: cortes.dat\n",
    "ARQUIVO DE CABECALHO CORTES : cortesh.dat\n",
    "ARQUIVO P/DESPACHO HIDROTERM: newdesp.dat\n",
    "ARQUIVO DE ESTADOS CORTES   : cortese.dat\n",
    "ARQUIVO DE ENERGIAS FORWARD : energiaf.dat\n",
    "ARQUIVO DE RESTRICOES SAR   : rsar.dat\n",
    "ARQUIVO DE CABECALHO SAR    : rsarh.dat\n",
    "ARQUIVO DE INDICE SAR       : rsari.dat\n",
    "ARQUIVO LISTAGEM CORTES     : nwlistcf.rel\n",
    "ARQUIVO LISTAGEM ESTADOS FCF: estados.rel\n",
    "ARQUIVO LISTAGEM SAR        : rsar.rel\n",
    "ARQUIVO DE ENERGIAS X FORW  : energiaxf.dat\n",
    "ARQUIVO DE VAZAO FORWARD    : vazaof.dat\n",
    "ARQUIVO DE VAZAO X FORWARD  : vazaoxf.dat\n",
]



import pandas as pd  # type: ignore

from cfinterface.files.sectionfile import SectionFile

from inewave.nwlistop.modelos.mediasmerc import TabelaMediasmerc


class Mediasmerc(SectionFile):
    """
    Armazena os dados das saídas referentes às médias de diversas variáveis
    agrupadas por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `MEDIAS-MERC.CSV`.

    """

    SECTIONS = [TabelaMediasmerc]

    @property
    def valores(self) -> pd.DataFrame:
        """
        Tabela com os valores de médias para as variáveis dos submercados.

        - estagio (`int`)
        - codigo_submercado (`int`)
        - <variavel_1> (`float`)
        - ...
        - <variavel_n> (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(TabelaMediasmerc)
        if isinstance(b, TabelaMediasmerc):
            return b.data
        return None



from inewave.newave.modelos.newavetim import BlocoTemposEtapasTim

from cfinterface.files.blockfile import BlockFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Newavetim(BlockFile):
    """
    Armazena os dados de saída do NEWAVE referentes aos
    tempos de execução do programa.

    """

    T = TypeVar("T")

    BLOCKS = [BlocoTemposEtapasTim]

    @property
    def tempos_etapas(self) -> Optional[pd.DataFrame]:
        """
        Tempos de execução do modelo por etapa do modelo.

        - etapa (`str`)
        - tempo (`timedelta`)

        :return: Os tempos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoTemposEtapasTim)
        if isinstance(b, BlocoTemposEtapasTim):
            return b.data
        return None



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (  # noqa
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.cbombsin import CBombAnos


class Cbombsin(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes ao consumo de energia para
    bombeamento para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `cbombsin.out`.

    """

    BLOCKS = [
        CBombAnos,
    ]



# Rotinas de testes associadas ao arquivo eolica-cadastro.csv do NEWAVE
from datetime import datetime
from inewave.libs.modelos.eolica import (
    RegistroPEECadastro,
    RegistroPEEPotenciaInstaladaPeriodo,
    RegistroPEEConfiguracaoPeriodo,
    RegistroPEEFTE,
    RegistroPEEGeracaoPatamar,
    RegistroHistoricoVentoHorizonte,
    RegistroHistoricoVento,
    RegistroPostoVentoCadastro,
    RegistroPEEPostoVento,
    RegistroPEESubmercado,
)

from inewave.libs import Eolica

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.eolica import (
    MockRegistroPEECadastro,
    MockRegistroPEEPotenciaInstaladaPeriodo,
    MockEolica,
    MockRegistroPEEConfiguracaoPeriodo,
    MockRegistroPEEFTE,
    MockRegistroPEEGeracaoPatamar,
    MockRegistroHistoricoVentoHorizonte,
    MockRegistroHistoricoVento,
    MockRegistroPostoCadastro,
    MockRegistroPEEPosto,
    MockRegistroPEESubmercado,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_registro_pee_cadastro():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPEECadastro))
    r = RegistroPEECadastro()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "NEInterior"]
    assert r.codigo_pee == 1
    r.codigo_pee = 2
    assert r.nome_pee == "NEInterior"
    r.nome_pee = "NEInterior"


def test_registro_pee_potencia_instalada_periodo():
    m: MagicMock = mock_open(
        read_data="".join(MockRegistroPEEPotenciaInstaladaPeriodo)
    )
    r = RegistroPEEPotenciaInstaladaPeriodo()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(year=2021, month=1, day=1),
        datetime(year=2030, month=12, day=1),
        6058.890,
    ]
    assert r.codigo_pee == 1
    r.codigo_pee = 2
    assert r.periodo_inicial == datetime(year=2021, month=1, day=1)
    r.periodo_inicial = datetime(year=2021, month=2, day=1)
    assert r.periodo_final == datetime(year=2030, month=12, day=1)
    r.periodo_final = datetime(year=2021, month=2, day=1)
    assert r.potencia_instalada == 6058.89
    r.potencia_instalada = 1000.0


def test_registro_pee_configuracao_periodo():
    m: MagicMock = mock_open(
        read_data="".join(MockRegistroPEEConfiguracaoPeriodo)
    )
    r = RegistroPEEConfiguracaoPeriodo()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(2021, 1, 1),
        datetime(2030, 12, 1),
        "centralizado",
    ]
    assert r.codigo_pee == 1
    r.codigo_pee = 5
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2030, 12, 1)
    r.data_fim = datetime(2030, 11, 1)
    assert r.estado_operacao == "centralizado"
    r.estado_operacao = "fixo"


def test_registro_peefte():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPEEFTE))
    r = RegistroPEEFTE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(2021, 1, 1),
        datetime(2030, 12, 1),
        -0.14454132687670500,
        0.10904637648150100,
    ]
    assert r.codigo_pee == 1
    r.codigo_pee = 5
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2030, 12, 1)
    r.data_fim = datetime(2030, 11, 1)
    assert r.coeficiente_linear == -0.14454132687670500
    r.coeficiente_linear = -0.5
    assert r.coeficiente_angular == 0.10904637648150100
    r.coeficiente_angular = 0.5


def test_registro_pee_geracao_patamar():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPEEGeracaoPatamar))
    r = RegistroPEEGeracaoPatamar()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2021, 1, 1), datetime(2021, 1, 1), 1, 0.88]
    assert r.codigo_pee == 1
    r.codigo_pee = 2
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2022, 1, 1)
    assert r.data_fim == datetime(2021, 1, 1)
    r.data_fim = datetime(2022, 1, 1)
    assert r.indice_patamar == 1
    r.indice_patamar = 2
    assert r.profundidade == 0.88
    r.profundidade = 2.0


def test_registro_historico_vento_horizonte_eolicahistorico():
    m: MagicMock = mock_open(
        read_data="".join(MockRegistroHistoricoVentoHorizonte)
    )
    r = RegistroHistoricoVentoHorizonte()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [datetime(1979, 1, 1), datetime(2016, 1, 1)]
    assert r.data_inicio == datetime(1979, 1, 1)
    r.data_inicio = datetime(1980, 1, 1)
    assert r.data_fim == datetime(2016, 1, 1)
    r.data_fim = datetime(2018, 1, 1)


def test_registro_historico_vento_eolicahistorico():
    m: MagicMock = mock_open(read_data="".join(MockRegistroHistoricoVento))
    r = RegistroHistoricoVento()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(1979, 1, 1), datetime(1979, 2, 1), 4.05, 1.0]
    assert r.codigo_posto == 1
    r.codigo_posto = 2
    assert r.data_inicio == datetime(1979, 1, 1)
    r.data_inicio = datetime(1980, 1, 1)
    assert r.data_fim == datetime(1979, 2, 1)
    r.data_fim = datetime(1980, 2, 1)
    assert r.velocidade == 4.05
    r.velocidade = 5.0
    assert r.direcao == 1.0
    r.direcao = 0.0


def test_registro_posto_cadastro_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPostoCadastro))
    r = RegistroPostoVentoCadastro()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "NEInterior"]
    assert r.codigo_posto == 1
    r.codigo_posto = 1
    assert r.nome_posto == "NEInterior"
    r.nome_posto = "NEInterior"


def test_registro_pee_posto_eolicaposto():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPEEPosto))
    r = RegistroPEEPostoVento()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1]
    assert r.codigo_posto == 1
    r.codigo_posto = 2
    assert r.codigo_pee == 1
    r.codigo_pee = 2


def test_registro_pee_submercado_eolicasubmercado():
    m: MagicMock = mock_open(read_data="".join(MockRegistroPEESubmercado))
    r = RegistroPEESubmercado()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 3]
    assert r.codigo_pee == 1
    r.codigo_pee = 2
    assert r.codigo_submercado == 3
    r.codigo_submercado = 1


def test_atributos_encontrados():
    m: MagicMock = mock_open(read_data="".join(MockEolica))
    with patch("builtins.open", m):
        e = Eolica.read(ARQ_TESTE)
        assert len(e.pee_cad()) > 0
        assert len(e.pee_config_per()) > 0
        assert len(e.pee_fpvp_lin_pu_per()) > 0
        assert len(e.pee_ger_prof_per_pat()) > 0
        assert len(e.posto_vento_cad()) > 0
        assert len(e.pee_posto()) > 0
        assert len(e.pee_subm()) > 0


def test_eq_eolica():
    m: MagicMock = mock_open(read_data="".join(MockEolica))
    with patch("builtins.open", m):
        cf1 = Eolica.read(ARQ_TESTE)
        cf2 = Eolica.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_eolica():
    m: MagicMock = mock_open(read_data="".join(MockEolica))
    with patch("builtins.open", m):
        cf1 = Eolica.read(ARQ_TESTE)
        cf2 = Eolica.read(ARQ_TESTE)
        cf2.data.remove(cf1.pee_cad()[0])
        assert cf1 != cf2


def test_leitura_escrita_eolica():
    m_leitura: MagicMock = mock_open(read_data="".join(MockEolica))
    with patch("builtins.open", m_leitura):
        cf1 = Eolica.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Eolica.read(ARQ_TESTE)
        assert cf1 == cf2



from cfinterface.components.block import Block
from cfinterface.components.line import Line

from inewave.config import (
    MESES_DF,
    MAX_PATAMARES,
    MAX_SERIES_SINTETICAS,
    MAX_UTES,
)
from inewave._utils.formatacao import formata_df_meses_para_datas_nwlistop

from typing import IO, List
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class ValoresClasseTermicaSeriePatamar(Block):
    """
    Bloco com a informaçao de uma tabela por submercado, com
    entradas por classe térmica, série e patamar.
    """

    __slots__ = [
        "__linha",
        "__linha_ano",
        "__serie_atual",
        "__classe_atual",
        "__ano",
    ]

    BEGIN_PATTERN = "     ANO: "
    END_PATTERN = "  TOTAL "
    HEADER_LINE = Line([])
    DATA_LINE = Line([])

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha_ano = self.__class__.HEADER_LINE
        self.__linha = self.__class__.DATA_LINE

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ValoresClasseTermicaSeriePatamar):
            return False
        bloco: ValoresClasseTermicaSeriePatamar = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            cols = MESES_DF
            df = pd.DataFrame(tabela, columns=["classe", "serie"] + cols)
            df["ano"] = self.__ano
            df["patamar"] = patamares
            df = df[["ano", "classe", "serie", "patamar"] + cols]
            df = df.astype(
                {"classe": "int64", "serie": "int64", "ano": "int64"}
            )
            return formata_df_meses_para_datas_nwlistop(df)

        self.__ano = self.__linha_ano.read(file.readline())[0]
        file.readline()

        # Variáveis auxiliares
        self.__classe_atual = 0
        self.__serie_atual = 0
        tabela = np.zeros(
            (
                MAX_PATAMARES * MAX_SERIES_SINTETICAS * MAX_UTES,
                len(MESES_DF) + 2,
            )
        )
        patamares: List[str] = []
        i = 0
        intervalo_classes = False
        while True:
            linha = file.readline()
            if self.ends(linha) or len(linha) <= 1:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break

            if "  MEDIA        " in linha:
                intervalo_classes = True
                continue
            elif "  MAX          " in linha:
                intervalo_classes = False
                continue

            if intervalo_classes:
                continue
            dados = self.__linha.read(linha)
            if dados[0] is not None:
                self.__classe_atual = dados[0]
            if dados[1] is not None:
                self.__serie_atual = dados[1]
            tabela[i, 0] = self.__classe_atual
            tabela[i, 1] = self.__serie_atual
            patamares.append(dados[2])
            tabela[i, 2:] = dados[3:]
            i += 1



from inewave.nwlistop.modelos.blocos.ree import REE
from inewave.nwlistop.modelos.arquivos.arquivoree import ArquivoREE

from inewave.nwlistop.modelos.mevmin import MevminAnos


class Mevmin(ArquivoREE):
    """
    Armazena os dados das saídas referentes à meta de energia de vazão
    mínima por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `mevmin00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        REE,
        MevminAnos,
    ]



from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.floatfield import FloatField
from typing import IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class SecaoDadosVazaof(Section):
    """
    Registro com os dados das séries sintéticas de vazão existentes
    nos arquivos vazaofXXX.dat.
    """

    __slots__ = ["__linha"]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, SecaoDadosVazaof):
            return False
        bloco: SecaoDadosVazaof = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    def read(
        self,
        file: IO,
        numero_forwards: int = 200,
        numero_uhes: int = 164,
        numero_estagios: int = 60,
        numero_estagios_th: int = 12,
        *args,
        **kwargs,
    ):
        numero_registros = (
            (numero_estagios + numero_estagios_th)
            * numero_uhes
            * numero_forwards
        )
        self.__linha = Line(
            [
                FloatField(size=8, starting_position=8 * i)
                for i in range(numero_registros)
            ],
            storage="BINARY",
        )
        dados = self.__linha.read(file.read(self.__linha.size))
        indices_estagios = np.arange(
            1 - numero_estagios_th, numero_estagios + 1
        )
        estagios_df = np.repeat(
            indices_estagios, numero_forwards * numero_uhes
        )
        uhes_df = np.tile(
            np.repeat(np.arange(1, numero_uhes + 1), numero_forwards),
            numero_estagios + numero_estagios_th,
        )
        forwards_df = np.tile(
            np.arange(1, numero_forwards + 1),
            numero_uhes * (numero_estagios + numero_estagios_th),
        )
        df = pd.DataFrame(
            data={
                "estagio": estagios_df,
                "uhe": uhes_df,
                "serie": forwards_df,
                "valor": dados,
            }
        )
        self.data = df

    def write(self, file: IO, *args, **kwargs):
        dados = self.data["valor"].to_numpy()
        linha = Line(
            [
                FloatField(size=8, starting_position=8 * i)
                for i in range(len(dados))
            ],
            storage="BINARY",
        )
        file.write(linha.write(dados))



from inewave.libs.usinas_hidreletricas import UsinasHidreletricas
from inewave.libs.usinas_hidreletricas import (
    HidreletricaCurvaJusante,
    HidreletricaCurvaJusantePolinomioPorPartes,
    HidreletricaCurvaJusantePolinomioPorPartesSegmento,
    HidreletricaCurvaJusanteAfogamentoExplicitoUsina,
    HidreletricaCurvaJusanteAfogamentoExplicitoPadrao,
    HidreletricaProdutibilidadeEspecificaGrade,
    HidreletricaPerdaHidraulicaGrade,
    EstacaoBombeamento,
    EstacaoBombeamentoLimitesPeriodoPatamar,
    EstacaoBombeamentoSubmercado,
    VolumeReferencialTipoPadrao,
    VolumeReferencialPeriodo,
)
from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch
from datetime import datetime
from tests.mocks.arquivos.usinas_hidreletricas import (
    MockHidreletricaCurvaJusante,
    MockHidreletricaCurvaJusantePolinomio,
    MockHidreletricaCurvaJusantePolinomioSegmento,
    MockHidreletricaCurvaJusanteAfogamentoExplicitoUsina,
    MockHidreletricaCurvaJusanteAfogamentoExplicitoPadrao,
    MockHidreletricaProdutibilidadeEspecificaGrade,
    MockHidreletricaPerdaHidraulicaGrade,
    MockEstacaoBombeamento,
    MockEstacaoBombeamentoLimitesPeriodoPatamar,
    MockEstacaoBombeamentoSubmercado,
    MockVolumeReferencialPadrao,
    MockVolumeReferencialPeriodo,
    MockUsinasHidreletricas,
)

ARQ_TESTE = "./tests/__init__.py"


def test_df_hidreletrica_curvajusante():
    m: MagicMock = mock_open(read_data="".join(MockHidreletricaCurvaJusante))
    with patch("builtins.open", m):
        polinjus = UsinasHidreletricas.read(ARQ_TESTE)
        df_curvajusante = polinjus.hidreletrica_curvajusante(df=True)
        assert df_curvajusante.at[0, "codigo_usina"] == 1
        assert df_curvajusante.at[0, "indice_familia"] == 1
        assert df_curvajusante.at[0, "nivel_montante_referencia"] == 885.3052


def test_registro_hidreletrica_curvajusante():
    m: MagicMock = mock_open(read_data="".join(MockHidreletricaCurvaJusante))
    r = HidreletricaCurvaJusante()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1, 885.3052]
    assert r.codigo_usina == 1
    r.codigo_usina = 0
    assert r.codigo_usina == 0
    assert r.indice_familia == 1
    r.indice_familia = 0
    assert r.indice_familia == 0
    assert r.nivel_montante_referencia == 885.3052
    r.nivel_montante_referencia = 0
    assert r.nivel_montante_referencia == 0


def test_df_hidreletrica_curvajusante_polinomio():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaCurvaJusantePolinomio)
    )
    with patch("builtins.open", m):
        polinjus = UsinasHidreletricas.read(ARQ_TESTE)
        df_curvajusante_polinomio = (
            polinjus.hidreletrica_curvajusante_polinomio(df=True)
        )
        assert df_curvajusante_polinomio.at[0, "codigo_usina"] == 1
        assert df_curvajusante_polinomio.at[0, "indice_familia"] == 1
        assert df_curvajusante_polinomio.at[0, "numero_polinomios"] == 2


def test_registro_hidreletrica_curvajusante_polinomio():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaCurvaJusantePolinomio)
    )
    r = HidreletricaCurvaJusantePolinomioPorPartes()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1, 2]
    assert r.codigo_usina == 1
    r.codigo_usina = 0
    assert r.codigo_usina == 0
    assert r.indice_familia == 1
    r.indice_familia = 0
    assert r.indice_familia == 0
    assert r.numero_polinomios == 2
    r.numero_polinomios = 0
    assert r.numero_polinomios == 0


def test_df_hidreletrica_curvajusante_polinomio_segmento():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaCurvaJusantePolinomioSegmento)
    )
    with patch("builtins.open", m):
        polinjus = UsinasHidreletricas.read(ARQ_TESTE)
        df_curvajusante_polinomio_segmento = (
            polinjus.hidreletrica_curvajusante_polinomio_segmento(df=True)
        )
        assert df_curvajusante_polinomio_segmento.at[0, "codigo_usina"] == 1
        assert df_curvajusante_polinomio_segmento.at[0, "indice_familia"] == 1
        assert (
            df_curvajusante_polinomio_segmento.at[0, "indice_polinomio"] == 1
        )
        assert (
            df_curvajusante_polinomio_segmento.at[
                0, "limite_inferior_vazao_jusante"
            ]
            == 0.0
        )
        assert (
            df_curvajusante_polinomio_segmento.at[
                0, "limite_superior_vazao_jusante"
            ]
            == 408.6490
        )
        assert (
            df_curvajusante_polinomio_segmento.at[0, "coeficiente_a0"]
            == 885.3052
        )
        assert (
            round(
                df_curvajusante_polinomio_segmento.at[0, "coeficiente_a1"], 4
            )
            == 0.0000
        )
        assert (
            round(
                df_curvajusante_polinomio_segmento.at[0, "coeficiente_a2"], 4
            )
            == 0.0000
        )
        assert (
            round(
                df_curvajusante_polinomio_segmento.at[0, "coeficiente_a3"], 4
            )
            == 0.0000
        )
        assert (
            round(
                df_curvajusante_polinomio_segmento.at[0, "coeficiente_a4"], 4
            )
            == 0.0000
        )


def test_registro_hidreletrica_curvajusante_polinomio_segmento():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaCurvaJusantePolinomioSegmento)
    )
    r = HidreletricaCurvaJusantePolinomioPorPartesSegmento()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        1,
        1,
        0,
        408.649,
        0.88530520000000e03,
        -0.31521360000000e-17,
        0.19686530000000e-04,
        -0.25518040000000e-07,
        0.00000000000000e01,
    ]
    assert r.codigo_usina == 1
    r.codigo_usina = 0
    assert r.codigo_usina == 0
    assert r.indice_familia == 1
    r.indice_familia = 0
    assert r.indice_familia == 0
    assert r.indice_polinomio == 1
    r.indice_polinomio = 0
    assert r.indice_polinomio == 0
    assert r.limite_inferior_vazao_jusante == 0
    r.limite_inferior_vazao_jusante = -1
    assert r.limite_inferior_vazao_jusante == -1
    assert r.limite_superior_vazao_jusante == 408.649
    r.limite_superior_vazao_jusante = -1
    assert r.limite_superior_vazao_jusante == -1

    assert r.coeficiente_a0 == 0.88530520000000e03
    r.coeficiente_a0 = -1
    assert r.coeficiente_a0 == -1
    assert r.coeficiente_a1 == -0.31521360000000e-17
    r.coeficiente_a1 = -1
    assert r.coeficiente_a1 == -1
    assert r.coeficiente_a2 == 0.19686530000000e-04
    r.coeficiente_a2 = -1
    assert r.coeficiente_a2 == -1
    assert r.coeficiente_a3 == -0.25518040000000e-07
    r.coeficiente_a3 = -1
    assert r.coeficiente_a3 == -1
    assert r.coeficiente_a4 == 0.00000000000000e01
    r.coeficiente_a4 = -1
    assert r.coeficiente_a4 == -1


def test_df_polinjus_hidreletrica_curvajusante_afogamentoexplicito_usina():
    m: MagicMock = mock_open(read_data="".join(MockUsinasHidreletricas))
    with patch("builtins.open", m):
        polinjus = UsinasHidreletricas.read(ARQ_TESTE)
        df_curvajusante_afogamentoexplicito_usina = (
            polinjus.hidreletrica_curvajusante_afogamentoexplicito_usina(
                df=True
            )
        )
        assert (
            df_curvajusante_afogamentoexplicito_usina.at[2, "codigo_usina"]
            == 54
        )
        assert (
            df_curvajusante_afogamentoexplicito_usina.at[
                2, "considera_afogamento"
            ]
            == "nao"
        )


def test_registro_polinjus_hidreletrica_curvajusante_afogamentoexplicito_usina():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaCurvaJusanteAfogamentoExplicitoUsina)
    )
    r = HidreletricaCurvaJusanteAfogamentoExplicitoUsina()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [22, "sim"]
    assert r.codigo_usina == 22
    r.codigo_usina = 0
    assert r.codigo_usina == 0
    assert r.considera_afogamento == "sim"
    r.considera_afogamento = "nao"
    assert r.considera_afogamento == "nao"


def test_registro_polinjus_hidreletrica_curvajusante_afogamentoexplicito_padrao():
    m: MagicMock = mock_open(
        read_data="".join(
            MockHidreletricaCurvaJusanteAfogamentoExplicitoPadrao
        )
    )
    r = HidreletricaCurvaJusanteAfogamentoExplicitoPadrao()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == ["nao"]
    assert r.considera_afogamento == "nao"
    r.considera_afogamento = "sim"
    assert r.considera_afogamento == "sim"


def test_registro_hidreletrica_produtibilidade_especifica_grade():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaProdutibilidadeEspecificaGrade)
    )
    r = HidreletricaProdutibilidadeEspecificaGrade()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [7, 0, 10, 0.012345]
    assert r.codigo_usina == 7
    r.codigo_usina = 2
    assert r.codigo_usina == 2
    assert r.altura_queda_liquida == 0.0
    r.altura_queda_liquida = 1.0
    assert r.altura_queda_liquida == 1.0
    assert r.turbinamento == 10.0
    r.turbinamento = 1.0
    assert r.turbinamento == 1.0
    assert r.produtibilidade_especifica == 0.012345
    r.produtibilidade_especifica = 1.5
    assert r.produtibilidade_especifica == 1.5


def test_registro_hidreletrica_perda_hidraulica_grade():
    m: MagicMock = mock_open(
        read_data="".join(MockHidreletricaPerdaHidraulicaGrade)
    )
    r = HidreletricaPerdaHidraulicaGrade()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 10.0, 1.0]
    assert r.codigo_usina == 1
    r.codigo_usina = 2
    assert r.codigo_usina == 2
    assert r.turbinamento == 10.0
    r.turbinamento = 1.0
    assert r.turbinamento == 1.0
    assert r.perda_hidraulica == 1.0
    r.perda_hidraulica = 1.5
    assert r.perda_hidraulica == 1.5


def test_registro_estacao_bombeamento():
    m: MagicMock = mock_open(read_data="".join(MockEstacaoBombeamento))
    r = EstacaoBombeamento()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "Sta Cecilia", 125, 181, 0.20, 181]
    assert r.codigo_estacao == 1
    r.codigo_estacao = 2
    assert r.codigo_estacao == 2
    assert r.nome_estacao == "Sta Cecilia"
    r.nome_estacao = "Sto Cecilia"
    assert r.nome_estacao == "Sto Cecilia"
    assert r.codigo_usina_origem == 125
    r.codigo_usina_origem = 124
    assert r.codigo_usina_origem == 124
    assert r.codigo_usina_destino == 181
    r.codigo_usina_destino = 180
    assert r.codigo_usina_destino == 180
    assert r.consumo_estacao == 0.2
    r.consumo_estacao = 0.5
    assert r.consumo_estacao == 0.5
    assert r.bombeamento_maximo == 181
    r.bombeamento_maximo = 180
    assert r.bombeamento_maximo == 180


def test_registro_estacao_bombeamento_limites_periodo_patamar():
    m: MagicMock = mock_open(
        read_data="".join(MockEstacaoBombeamentoLimitesPeriodoPatamar)
    )
    r = EstacaoBombeamentoLimitesPeriodoPatamar()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2023, 3, 1), None, 1, 0, 181]
    assert r.codigo_estacao == 1
    r.codigo_estacao = 2
    assert r.codigo_estacao == 2
    assert r.data_inicio == datetime(2023, 3, 1)
    r.data_inicio = datetime(2023, 2, 1)
    assert r.data_inicio == datetime(2023, 2, 1)
    assert r.data_fim is None
    r.data_fim = datetime(2023, 3, 2)
    assert r.data_fim == datetime(2023, 3, 2)
    assert r.patamar == 1
    r.patamar = 3
    assert r.patamar == 3
    assert r.limite_inferior == 0
    r.limite_inferior = 3
    assert r.limite_inferior == 3
    assert r.limite_superior == 181
    r.limite_superior = 151
    assert r.limite_superior == 151


def test_registro_estacao_bombeamento_submercado():
    m: MagicMock = mock_open(
        read_data="".join(MockEstacaoBombeamentoSubmercado)
    )
    r = EstacaoBombeamentoSubmercado()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1]
    assert r.codigo_estacao == 1
    r.codigo_estacao = 2
    assert r.codigo_estacao == 2
    assert r.codigo_submercado == 1
    r.codigo_submercado = 2
    assert r.codigo_submercado == 2


def test_registro_volume_referencial_tipo_padrao():
    m: MagicMock = mock_open(read_data="".join(MockVolumeReferencialPadrao))
    r = VolumeReferencialTipoPadrao()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1]
    assert r.tipo_referencia == 1
    r.tipo_referencia = 2
    assert r.tipo_referencia == 2


def test_registro_volume_referencial_periodo():
    m: MagicMock = mock_open(read_data="".join(MockVolumeReferencialPeriodo))
    r = VolumeReferencialPeriodo()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [4, datetime(2023, 4, 1), None, 266.0]
    assert r.codigo_usina == 4
    r.codigo_usina = 5
    assert r.codigo_usina == 5
    assert r.data_inicio == datetime(2023, 4, 1)
    r.data_inicio = datetime(2023, 5, 1)
    assert r.data_inicio == datetime(2023, 5, 1)
    assert r.data_fim is None
    r.data_fim = datetime(2023, 6, 1)
    assert r.data_fim == datetime(2023, 6, 1)
    assert r.volume_referencia == 266.0
    r.volume_referencia = 0.0
    assert r.volume_referencia == 0.0


def test_atributos_encontrados():
    m: MagicMock = mock_open(read_data="".join(MockUsinasHidreletricas))
    with patch("builtins.open", m):
        uhes = UsinasHidreletricas.read(ARQ_TESTE)
        assert uhes.hidreletrica_curvajusante() is not None
        assert uhes.hidreletrica_curvajusante_polinomio() is not None
        assert uhes.hidreletrica_curvajusante_polinomio_segmento() is not None
        assert (
            uhes.hidreletrica_curvajusante_afogamentoexplicito_usina()
            is not None
        )
        assert (
            uhes.hidreletrica_curvajusante_afogamentoexplicito_padrao()
            is not None
        )
        assert uhes.hidreletrica_produtibilidade_especifica_grade() is not None
        assert uhes.hidreletrica_perda_hidraulica_grade() is not None
        assert uhes.estacao_bombeamento() is not None
        assert uhes.estacao_bombeamento_limites_periodo_patamar() is not None
        assert uhes.estacao_bombeamento_submercado() is not None
        assert uhes.volume_referencial_tipo_padrao() is not None
        assert uhes.volume_referencial_periodo() is not None


def test_eq():
    m: MagicMock = mock_open(read_data="".join(MockUsinasHidreletricas))
    with patch("builtins.open", m):
        log1 = UsinasHidreletricas.read(ARQ_TESTE)
        log2 = UsinasHidreletricas.read(ARQ_TESTE)
        assert log1 == log2


def test_neq():
    m: MagicMock = mock_open(read_data="".join(MockUsinasHidreletricas))
    with patch("builtins.open", m):
        log1 = UsinasHidreletricas.read(ARQ_TESTE)
        log2 = UsinasHidreletricas.read(ARQ_TESTE)
        log1.hidreletrica_curvajusante_polinomio_segmento()[
            0
        ].codigo_usina = -1
        assert log1 != log2


def test_leitura_escrita_usinas_hidreletricas():
    m_leitura: MagicMock = mock_open(
        read_data="".join(MockUsinasHidreletricas)
    )
    with patch("builtins.open", m_leitura):
        cf1 = UsinasHidreletricas.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = UsinasHidreletricas.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousina import (
    ArquivoUsina,
)
from inewave.nwlistop.modelos.pivarmincr import PivarmAnos, PivarmAnos_v29_2


class Pivarmincr(ArquivoUsina):
    """
    Armazena os dados das saídas referentes aos valores da água
    incrementais por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `pivarmincr00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        PivarmAnos_v29_2,
    ]

    VERSIONS = {
        "28.12": [
            Usina,
            PivarmAnos,
        ],
        "29.2": [
            Usina,
            PivarmAnos_v29_2,
        ],
    }



from inewave.nwlistop.modelos.merclsin import MerclAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class Merclsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao mercado líquido
    de cada estágio em cada série.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `merclsin.out`.
    """

    BLOCKS = [
        MerclAnos,
    ]



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousina import (
    ArquivoUsina,
)
from inewave.nwlistop.modelos.viol_neg_evap import ViolNegEvapAnos


class ViolNegEvap(ArquivoUsina):
    """
    Armazena os dados das saídas referentes ao desvios negativos da
    evaporação da usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_neg_evap00x.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        ViolNegEvapAnos,
    ]



from inewave.newave.modelos.exph import BlocoUHEExph

from cfinterface.files.sectionfile import SectionFile
from cfinterface.components.section import Section
from typing import TypeVar, List, Type, Optional
import pandas as pd  # type: ignore


class Exph(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à expansão
    hidraulica do sistema.

    """

    T = TypeVar("T")

    SECTIONS: List[Type[Section]] = [BlocoUHEExph]

    @property
    def expansoes(self) -> Optional[pd.DataFrame]:
        """
        A tabela de expansões de máquinas das UHEs.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - data_inicio_enchimento (`datetime`)
        - duracao_enchimento (`int`)
        - volume_morto (`float`)
        - data_entrada_operacao (`datetime`)
        - potencia_instalada (`float`)
        - maquina_entrada (`int`)
        - conjunto_maquina_entrada (`int`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUHEExph)
        if isinstance(b, BlocoUHEExph):
            return b.data
        return None

    @expansoes.setter
    def expansoes(self, d: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUHEExph)
        if isinstance(b, BlocoUHEExph):
            b.data = d



from inewave.nwlistop.modelos.blocos.ree import REE
from inewave.nwlistop.modelos.arquivos.arquivoree import ArquivoREE

from inewave.nwlistop.modelos.vagua import VAAnos

from warnings import warn


class Vagua(ArquivoREE):
    """
    Armazena os dados das saídas referentes aos valores da água
    por REE.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `vagua00x.out`, onde x varia conforme o
    REE em questão.

    """

    BLOCKS = [
        REE,
        VAAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe ValorAgua no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from inewave.newave.modelos.dsvagua import BlocoDsvUHE

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Dsvagua(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    desvios de água por usina.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que podem ser modificadas através do arquivo `dsvagua.dat`.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoDsvUHE]

    @property
    def desvios(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os desvios de água por usina e por estágio.

        - codigo_usina (`int`)
        - data (`datetime`)
        - valor (`float`)
        - considera_desvio_usina_NC (`int`)
        - comentario (`str`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoDsvUHE)
        if isinstance(b, BlocoDsvUHE):
            return b.data
        return None

    @desvios.setter
    def desvios(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoDsvUHE)
        if isinstance(b, BlocoDsvUHE):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



from inewave.newave.modelos.patamar import (
    BlocoNumeroPatamares,
    BlocoDuracaoPatamar,
    BlocoCargaPatamar,
    BlocoIntercambioPatamarSubsistemas,
    BlocoUsinasNaoSimuladas,
)

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Patamar(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    patamares de carga por submercado.

    """

    T = TypeVar("T")

    SECTIONS = [
        BlocoNumeroPatamares,
        BlocoDuracaoPatamar,
        BlocoCargaPatamar,
        BlocoIntercambioPatamarSubsistemas,
        BlocoUsinasNaoSimuladas,
    ]

    @property
    def numero_patamares(self) -> Optional[int]:
        """
        O número de patamares utilizado no estudo.

        :return: O número de patamares como um inteiro
        :rtype: int | None
        """
        b = self.data.get_sections_of_type(BlocoNumeroPatamares)
        if isinstance(b, BlocoNumeroPatamares):
            return b.data
        return None

    @numero_patamares.setter
    def numero_patamares(self, n: int):
        b = self.data.get_sections_of_type(BlocoNumeroPatamares)
        if isinstance(b, BlocoNumeroPatamares):
            b.data = n

    @property
    def duracao_mensal_patamares(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a duração mensal de cada patamar no horizonte
        de estudo.

        - data (`datetime`)
        - patamar (`int`)
        - valor (`float`)

        :return: A duração por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoDuracaoPatamar)
        if isinstance(b, BlocoDuracaoPatamar):
            return b.data
        return None

    @duracao_mensal_patamares.setter
    def duracao_mensal_patamares(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoDuracaoPatamar)
        if isinstance(b, BlocoDuracaoPatamar):
            b.data = df

    @property
    def carga_patamares(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a carga em P.U. por patamar.

        - codigo_submercado (`int`)
        - data (`datetime`)
        - patamar (`int`)
        - valor (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoCargaPatamar)
        if isinstance(b, BlocoCargaPatamar):
            return b.data
        return None

    @carga_patamares.setter
    def carga_patamares(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoCargaPatamar)
        if isinstance(b, BlocoCargaPatamar):
            b.data = df

    @property
    def intercambio_patamares(self) -> Optional[pd.DataFrame]:
        """
        Tabela com a correção em P.U. do intercâmbio por patamar.

        - submercado_de (`int`)
        - submercado_para (`int`)
        - data (`datetime`)
        - patamar (`int`)
        - valor (`float`)

        :return: A carga por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoIntercambioPatamarSubsistemas)
        if isinstance(b, BlocoIntercambioPatamarSubsistemas):
            return b.data
        return None

    @intercambio_patamares.setter
    def intercambio_patamares(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoIntercambioPatamarSubsistemas)
        if isinstance(b, BlocoIntercambioPatamarSubsistemas):
            b.data = df

    @property
    def usinas_nao_simuladas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os fatores das usinas não simuladas em P.U.

        - codigo_submercado (`int`)
        - indice_bloco (`int`)
        - data (`datetime`)
        - patamar (`int`)
        - valor (`float`)

        :return: Os valores por mês em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUsinasNaoSimuladas)
        if isinstance(b, BlocoUsinasNaoSimuladas):
            return b.data
        return None

    @usinas_nao_simuladas.setter
    def usinas_nao_simuladas(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUsinasNaoSimuladas)
        if isinstance(b, BlocoUsinasNaoSimuladas):
            b.data = df



from cfinterface.components.floatfield import FloatField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.line import Line

from inewave.config import MESES_DF
from inewave.nwlistop.modelos.blocos.valoresserie import ValoresSerie


class CmargsAnos28(ValoresSerie):
    """
    Bloco com a informaçao do submercado associado aos valores de Custo
    Marginal de Operação.
    """

    __slots__ = []

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
        ]
        + [FloatField(8, 9 + 10 * i, 2) for i in range(len(MESES_DF))]  # type: ignore
    )


class CmargsAnos(ValoresSerie):
    """
    Bloco com a informaçao do submercado associado aos valores de Custo
    Marginal de Operação.
    """

    __slots__ = []

    HEADER_LINE = Line([IntegerField(4, 10)])
    DATA_LINE = Line(
        [  # type: ignore
            IntegerField(4, 2),
        ]
        + [FloatField(11, 9 + 11 * i, 2) for i in range(len(MESES_DF))]  # type: ignore
    )



from inewave.nwlistop.modelos.cdefsin import CdefAnos

from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)


class Cdefsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao custo de déficit
    de cada estágio em cada série.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cdefsin.out`.
    """

    BLOCKS = [
        CdefAnos,
    ]



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.depminuh import DepminAnos

from warnings import warn


class Depminuh(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes À violação de defluência
    mínima da usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `depminuh00x.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        DepminAnos,
    ]

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe ViolVazmin no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.evertsin import EvertAnos


class Evertsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes ao vertimento de reservatórios
    , para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas no `evertsin.out`.

    """

    BLOCKS = [
        EvertAnos,
    ]



from inewave.nwlistop.modelos.arquivos.arquivosubmercadopatamar import (
    ArquivoSubmercadoPatamar,
)
from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.cmarg import CmargsAnos, CmargsAnos27


class Cmarg(ArquivoSubmercadoPatamar):
    """
    Armazena os dados das saídas referentes aos custos marginais de operação
    por patamar, por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `cmarg00x.out`, onde x varia conforme o
    submercado em questão.

    """

    BLOCKS = [
        Submercado,
        CmargsAnos,
    ]

    VERSIONS = {
        "27": [
            Submercado,
            CmargsAnos27,
        ],
        "29.4.1": [Submercado, CmargsAnos],
    }



from inewave.newave.modelos.volref_saz import BlocoVolrefSaz

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class VolrefSaz(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    volumes de referência sazonais para as usinas.

    Esta classe lida com informações de entrada fornecidas ao NEWAVE e
    que são usadas para fins de cálculos da evaporação linear e ajustes
    da FPHA.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoVolrefSaz]

    @property
    def volumes(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os volumes de referência sazonais por UHE.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - mes (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoVolrefSaz)
        if isinstance(b, BlocoVolrefSaz):
            return b.data
        return None

    @volumes.setter
    def volumes(self, df: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoVolrefSaz)
        if isinstance(b, BlocoVolrefSaz):
            b.data = df



from inewave.newave.modelos.parpvaz import (
    BlocoSerieVazoesUHE,
    BlocoCorrelVazoesUHE,
    BlocoCorrelParcialVazoesUHE,
    BlocoOrdemModeloUHE,
    BlocoCoeficientesModeloUHE,
    BlocoSerieRuidosUHE,
    BlocoCorrelRuidosUHE,
    BlocoCorrelEspacialAnualMensalUHE,
)

from inewave.newave import Parpvaz


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.parpvaz import (
    MockSeriesVazoesPARpvaz,
    MockCorrelogramoSerieVazoesPARpvaz,
    MockCorrelogramoParcialSerieVazoesPARpvaz,
    MockOrdemOriginalModeloPARpvaz,
    MockCoeficientesModeloPARpvaz,
    MockSerieRuidosPARpvaz,
    MockCorrelogramoSerieRuidosPARpvaz,
    MockCorrelacaoEspacialPARpvaz,
    MockPARpvaz,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_series_vazoes_uhe():
    m: MagicMock = mock_open(read_data="".join(MockSeriesVazoesPARpvaz))
    b = BlocoSerieVazoesUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 1080
    assert b.data.shape[1] == 4
    assert b.data.iloc[0, 3] == 302.0
    assert b.data.iloc[-1, -1] == 176.0


def test_correlacao_series_vazoes_uhe():
    m: MagicMock = mock_open(
        read_data="".join(MockCorrelogramoSerieVazoesPARpvaz)
    )
    b = BlocoCorrelVazoesUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 132
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 2] == 0.45003
    assert b.data.iloc[-1, -1] == 0.2489


def test_correlacao_parcial_series_vazoes_uhe():
    m: MagicMock = mock_open(
        read_data="".join(MockCorrelogramoParcialSerieVazoesPARpvaz)
    )
    b = BlocoCorrelParcialVazoesUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 132
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 2] == 0.45003
    assert b.data.iloc[-1, -1] == 0.00987


def test_ordem_modelo_uhe():
    m: MagicMock = mock_open(read_data="".join(MockOrdemOriginalModeloPARpvaz))
    b = BlocoOrdemModeloUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 12
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 2] == 6
    assert b.data.iloc[-1, -1] == 6


def test_coeficientes_modelo_uhe():
    m: MagicMock = mock_open(read_data="".join(MockCoeficientesModeloPARpvaz))
    b = BlocoCoeficientesModeloUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)
    assert b.data.shape[0] == 4
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 2] == 0.450
    assert b.data.iloc[1, 2] == 0.0
    assert b.data.iloc[2, 2] == 0.727
    assert b.data.iloc[3, 2] == 0.0


def test_serie_ruidos_uhe():
    m: MagicMock = mock_open(read_data="".join(MockSerieRuidosPARpvaz))
    b = BlocoSerieRuidosUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 1080
    assert b.data.shape[1] == 4
    assert b.data.iloc[0, 3] == 0.0
    assert b.data.iloc[-1, -1] == -0.0752


def test_correl_serie_ruidos_uhe():
    m: MagicMock = mock_open(
        read_data="".join(MockCorrelogramoSerieRuidosPARpvaz)
    )
    b = BlocoCorrelRuidosUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 132
    assert b.data.shape[1] == 3
    assert b.data.iloc[0, 2] == -0.06325
    assert b.data.iloc[-1, -1] == 0.00761


def test_correl_espacial_anual_mensal():
    m: MagicMock = mock_open(read_data="".join(MockCorrelacaoEspacialPARpvaz))
    b = BlocoCorrelEspacialAnualMensalUHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data.shape[0] == 17
    assert b.data.shape[1] == 15
    assert b.data.iloc[0, 2] == 0.55
    assert b.data.iloc[-1, -1] == 0.52


def test_atributos_encontrados_parpvaz():
    m: MagicMock = mock_open(read_data="".join(MockPARpvaz))
    with patch("builtins.open", m):
        parp = Parpvaz.read(ARQ_TESTE)
        assert parp.series_vazoes_uhe is not None
        assert parp.correlacao_series_vazoes_uhe is not None
        assert parp.correlacao_parcial_series_vazoes_uhe is not None
        assert parp.ordem_original_modelo is not None
        assert parp.ordem_final_modelo is not None
        assert parp.coeficientes is not None
        assert parp.series_ruido_uhe is not None
        assert parp.correlacao_series_ruidos_uhe is not None
        assert parp.correlacao_espacial_anual_mensal is not None


def test_atributos_nao_encontrados_parpvaz():
    m: MagicMock = mock_open(read_data="".join(MockSerieRuidosPARpvaz))
    with patch("builtins.open", m):
        parp = Parpvaz.read(ARQ_TESTE)
        assert parp.series_vazoes_uhe is None
        assert parp.correlacao_series_vazoes_uhe is None
        assert parp.correlacao_parcial_series_vazoes_uhe is None
        assert parp.ordem_original_modelo is None
        assert parp.ordem_final_modelo is None
        assert parp.coeficientes is None
        assert parp.series_ruido_uhe is None
        assert parp.correlacao_series_ruidos_uhe is None
        assert parp.correlacao_espacial_anual_mensal is None


def test_eq_parpvaz():
    m: MagicMock = mock_open(read_data="".join(MockPARpvaz))
    with patch("builtins.open", m):
        parp1 = Parpvaz.read(ARQ_TESTE)
        parp2 = Parpvaz.read(ARQ_TESTE)
        assert parp1 == parp2


def test_neq_parpvaz():
    m: MagicMock = mock_open(read_data="".join(MockPARpvaz))
    with patch("builtins.open", m):
        parp1 = Parpvaz.read(ARQ_TESTE)
        parp2 = Parpvaz.read(ARQ_TESTE)
        parp2.series_vazoes_uhe.iloc[0, 0] = -1
        assert parp1 != parp2



MockBlocoVarreduraShist = [
    "VARREDURA    E   ANO INICIO VARREDURA\n",
    "XXX.XXX.\n",
    "  1 1932\n",
]

MockBlocoSeriesSimulacaoShist = [
    "SERIES PARA SIMULACAO\n",
    "XXXX\n",
    "9999\n",
]

MockShist = MockBlocoVarreduraShist + MockBlocoSeriesSimulacaoShist



from inewave.newave.modelos.blocos.versaomodelo import VersaoModelo
from inewave.newave.modelos.nwv_eco_evap import TabelaEcoEvap

from inewave.newave.modelos.arquivoscsv.arquivocsv import ArquivoCSV
from typing import Optional
import pandas as pd  # type: ignore

from warnings import warn


class NwvEcoEvap(ArquivoCSV):
    """
    Arquivo com o eco dos dados da evaporação linear do NEWAVE.
    """

    BLOCKS = [VersaoModelo, TabelaEcoEvap]

    @property
    def tabela(self) -> Optional[pd.DataFrame]:
        """
        A tabela de dados que está contida no arquivo.

        - periodo (`int`)
        - codigo_usina (`int`)
        - nome_usina (`str`)
        - volume_referencia_hm3 (`float`)
        - evaporacao_referencia_hm3 (`float`)
        - coeficiente_evaporacao_mm_mes (`int`)
        - flag_evaporacao (`int`)
        - evaporacao_linear (`int`)
        - tipo_volume_referencia (`int`)

        :return: A tabela como um dataframe
        :rtype: pd.DataFrame | None
        """
        return self._tabela()

    def __init__(self, data=...) -> None:
        warn(
            "Esta classe é relativa a um arquivo que não é mais suportado."
            + " Utilize a classe EvapEco no lugar.",
            DeprecationWarning,
        )
        super().__init__(data)



from inewave.nwlistop.dlpptbmax import Dlpptbmax

from datetime import datetime
from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.dlpptbmax import MockDLPPtbmax

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_dlpptbmax():
    m: MagicMock = mock_open(read_data="".join(MockDLPPtbmax))
    with patch("builtins.open", m):
        n = Dlpptbmax.read(ARQ_TESTE)
        assert n.ree is not None
        assert n.ree == "SUDESTE"
        assert n.valores is not None
        assert n.valores.iloc[0, 0] == datetime(2020, 1, 1)
        assert n.valores.iloc[-1, -1] == 0.0


def test_atributos_nao_encontrados_dlpptbmax():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        n = Dlpptbmax.read(ARQ_TESTE)
        assert n.ree is None
        assert n.valores is None


def test_eq_dlpptbmax():
    m: MagicMock = mock_open(read_data="".join(MockDLPPtbmax))
    with patch("builtins.open", m):
        n1 = Dlpptbmax.read(ARQ_TESTE)
        n2 = Dlpptbmax.read(ARQ_TESTE)
        assert n1 == n2


# Não deve ter teste de diferença, visto que o atributo é
# implementado como Lazy Property.



"""
=====================================================
Geração de sensibilidades de mercado no sistema.dat
=====================================================
"""


# %%
# Para realizar a análise do sistema.dat, será utilizado o módulo plotly
from datetime import datetime
import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio

pio.templates.default = "ggplot2"


# %%
# O sistema.dat é o arquivo de entrada do modelo NEWAVE que contém informações
# sobre os submercados de energia. Em particular, são definidos os submercados
# e as curvas de mercado de energia esperado por estágio, para cada um.
# É uma análise comum a análise de sensibilidades em relação ao mercado de energia
# e à geração das usinas não simuladas e, por isso, será ilustrado um exemplo deste caso.
from inewave.newave import Sistema

arq_sistema = Sistema.read("./newave/sistema.dat")

# %%
# A definição dos submercados é acessível através do bloco que define os custos de déficit
arq_sistema.custo_deficit

# %%
# As informações de mercado de energia são reunidas em uma única propriedade
# e esta pode ser alterada livremente
arq_sistema.mercado_energia


# %%
# Será feito um gráfico de área empilhado. Para isso, serão geradas algumas variáveis auxiliares.
df = arq_sistema.mercado_energia
x = pd.date_range(
    df["data"].min(),
    df["data"].max(),
    freq="MS",
)

# %%
# Para a figura, são geradas as retas independentemente
fig = go.Figure()
for submercado in df["codigo_submercado"].unique():
    y = df.loc[df["codigo_submercado"] == submercado, "valor"]
    fig.add_trace(
        go.Scatter(
            x=x,
            y=y,
            mode="lines",
            stackgroup="one",
            name=str(submercado),
        )
    )

fig.update_xaxes(title="Data")
fig.update_yaxes(title="Mercado de Energia (MWmes)")
fig.update_layout(legend_title_text="Submercado")
fig

# %%
# É possível realizar edições livres na propriedade do arquivo, para geração de estudos
# de sensibilidades. Por exemplo, é possível aumentar a carga do submercado NORDESTE
# em 30% e conferir o efeito na operação com a execução do modelo.
arq_sistema.mercado_energia.loc[
    arq_sistema.mercado_energia["codigo_submercado"] == 3, "valor"
] *= 1.3

from io import StringIO

conteudo_sistema = StringIO()
arq_sistema.write(conteudo_sistema)
print(conteudo_sistema.getvalue())



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)

from inewave.nwlistop.modelos.mevminsin import MevminAnos


class Mevminsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes à meta de energia
    de vazão mínima para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `mevminsin.out`.

    """

    BLOCKS = [
        MevminAnos,
    ]
